#define MAX_KMALLOC_MEM (128*1024)
/*
    mv300 - AverMedia MV-300 driver version 0.0.1a

	Copyright (C) 1999 Kazhiro Mogi <ee69011@isc.meiji.ac.jp>

   based on
  
    buz - Iomega Buz driver version 1.0

    Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>

   based on

    buz.0.0.3 Copyright (C) 1998 Dave Perks <dperks@ibm.net>

   and

    bttv - Bt848 frame grabber driver

    Copyright (C) 1996,97,98 Ralph  Metzler (rjkm@thp.uni-koeln.de)
                           & Marcus Metzler (mocm@thp.uni-koeln.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>
#include <asm/spinlock.h>

#include <linux/videodev.h>

#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/i2c.h>
#include "buz.h"
#include "video_decoder.h"
#include "video_encoder.h"

#if defined(CONFIG_BIGPHYS_AREA)
#   include <linux/bigphysarea.h>
#endif

#define IRQ_MASK ( ZR36057_ISR_GIRQ0 | /* ZR36057_ISR_GIRQ1 | ZR36057_ISR_CodRepIRQ | */ ZR36057_ISR_JPEGRepIRQ )
#define GPIO_MASK 0xdf

/*
GPIO 0 = KS0127, Bt866, ZR36060 RESET
GPIO 1 = ZR36060 SLEEP
GPIO 2 = Bt866 SLAVE
GPIO 3 = ZR36060 FRAME
GPIO 4 = Bt866 SLEEP
GPIO 5 = Unknown
GPIO 6 = Unknown (Analog MUX?)
GPIO 7 = Unknown 

GIRQ0 signals a vertical sync of the video signal?
GIRQ1 signals that ZR36060's DATERR# line is asserted?

*/

#define MAJOR_VERSION 1    /* driver major version */
#define MINOR_VERSION 0    /* driver minor version */

#define MV300_NAME      "AverMedia MV-300"   /* name of the driver */

#define MV300_ERR       KERN_ERR     MV300_NAME
#define MV300_DEBUG     KERN_INFO    MV300_NAME
#define MV300_INFO      KERN_INFO    MV300_NAME
#define MV300_WARNING   KERN_WARNING MV300_NAME

#define DEBUG(x)           /* Debug driver */
#define IDEBUG(x)          /* Debug interrupt handler */
#define IOCTL_DEBUG(x)  


/* The parameters for this driver */

/*
   The video mem address of the video card.
   The driver has a little database for some videocards
   to determine it from there. If your video card is not in there
   you have either to give it to the driver as a parameter
   or set in in a VIDIOCSFBUF ioctl
 */

static unsigned long vidmem = 0;   /* Video memory base address */

/* Special purposes only: */

static int triton = 0;            /* 0=no, 1=yes */
static int natoma = 0;            /* 0=no, 1=yes */

/*
   Number and size of grab buffers for Video 4 Linux
   The vast majority of applications should not need more than 2,
   the very popular BTTV driver actually does ONLY have 2.
   Time sensitive applications might need more, the maximum
   is VIDEO_MAX_FRAME (defined in <linux/videodev.h>).

   The size is set so that the maximum possible request
   can be satisfied. Decrease  it, if bigphys_area alloc'd
   memory is low. If you don't have the bigphys_area patch,
   set it to 128 KB. Will you allow only to grab small
   images with V4L, but that's better than nothing.

   v4l_bufsize has to be given in KB !

*/

static int v4l_nbufs   = 2;
static int v4l_bufsize = 128; /* Everybody should be able to work with this setting */

/*
   Default input and video norm at startup of the driver.
*/

static int default_input = 0; /* 0=Composite, 1=S-VHS */
static int default_norm  = 0; /* 0=PAL, 1=NTSC */

MODULE_PARM(vidmem,        "i");
MODULE_PARM(triton,        "i");
MODULE_PARM(natoma,        "i");
MODULE_PARM(v4l_nbufs,     "i");
MODULE_PARM(v4l_bufsize,   "i");
MODULE_PARM(default_input, "i");
MODULE_PARM(default_norm , "i");

/* Anybody who uses more than four? */
#define MV300_MAX 4

static int mv300_num;         /* number of MV-300s in use */
static struct buz mv300s[MV300_MAX];

/* forward references */

static void v4l_fbuffer_free(struct buz * btv);
static void jpg_fbuffer_free(struct buz * btv);
static void mv300_feed_stat_com(struct buz * btv);



/*
 *   Allocate the V4L grab buffers
 *
 *   These have to be pysically contiguous.
 *   If v4l_bufsize <= MAX_KMALLOC_MEM we use kmalloc
 *   else we try to allocate them with bigphysarea_alloc_pages
 *   if the bigphysarea patch is present in the kernel,
 *   else we try to use high memory (if the user has bootet
 *   Linux with the necessary memory left over).
 */

// RJ: Test only - want to test MV300_USE_HIMEM even when CONFIG_BIGPHYS_AREA is defined
//#if defined(CONFIG_BIGPHYS_AREA)
//#undef CONFIG_BIGPHYS_AREA
//#endif

#define MV300_USE_HIMEM

#ifdef MV300_USE_HIMEM

static unsigned long get_high_mem(unsigned long size)
{
/*
 * Check if there is usable memory at the end of Linux memory
 * of at least size. Return the physical address of this memory,
 * return 0 on failure.
 *
 * The idea is from Alexandro Rubini's book "Linux device drivers".
 * The driver from him which is downloadable from O'Reilly's
 * web site misses the "virt_to_phys(high_memory)" part
 * (and therefore doesn't work at all - at least with 2.2.x kernels).
 *
 * It should be unnecessary to mention that THIS IS DANGEROUS,
 * if more than one driver at a time has the idea to use this memory!!!!
 */

   volatile unsigned char *mem;
   unsigned char c;
   unsigned long hi_mem_ph;
   unsigned long i;

   /* Map the high memory to user space */

   hi_mem_ph = virt_to_phys(high_memory);

   mem = ioremap(hi_mem_ph, size);
   if(!mem)
   {
      printk(MV300_ERR ": get_high_mem: ioremap failed\n");
      return 0;
   }

   for(i=0;i<size;i++)
   {
      /* Check if it is memory */
      c = i&0xff;
      mem[i] = c;
      if(mem[i] != c) break;
      c = 255-c;
      mem[i] = c;
      if(mem[i] != c) break;
      mem[i] = 0; /* zero out memory */

      /* give the kernel air to breath */
      if( (i&0x3ffff) == 0x3ffff) schedule();
   }

   iounmap((void *)mem);

   if(i!=size)
   {
      printk(MV300_ERR ": get_high_mem - requested %lu, avail %lu\n",size,i);
      return 0;
   }

   return hi_mem_ph;
}
#endif

static int v4l_fbuffer_alloc(struct buz * btv)
{
   int i, off;
   unsigned char *mem;
#if defined(MV300_USE_HIMEM) && !defined(CONFIG_BIGPHYS_AREA)
   unsigned long pmem = 0;
#endif

   for (i = 0; i<v4l_nbufs; i++)
   {
      if(btv->v4l_gbuf[i].fbuffer)
         printk(MV300_WARNING ": v4l_fbuffer_alloc: buffer %d allready allocated ???\n",i);

      if (v4l_bufsize <= MAX_KMALLOC_MEM)
      {
         /* Use kmalloc */

         mem = (unsigned char *)kmalloc(v4l_bufsize, GFP_KERNEL);
         if (mem==0)
         {
            printk(MV300_ERR ": kmalloc for V4L bufs failed\n");
            v4l_fbuffer_free(btv);
            return -ENOBUFS;
         }
         btv->v4l_gbuf[i].fbuffer      = mem;
         btv->v4l_gbuf[i].fbuffer_phys = virt_to_phys(mem);
         btv->v4l_gbuf[i].fbuffer_bus  = virt_to_bus (mem);
         for(off=0;off<v4l_bufsize;off+=PAGE_SIZE) mem_map_reserve(MAP_NR(mem+off));
         DEBUG(printk(MV300_INFO ": V4L frame %d mem 0x%x (bus: 0x%x=%d)\n",
                      i,mem,virt_to_bus(mem),virt_to_bus(mem));)
      }
      else
      {
#if defined(CONFIG_BIGPHYS_AREA)
         /* Use bigphysarea_alloc_pages */

         int n = (v4l_bufsize + PAGE_SIZE-1)/PAGE_SIZE;
         mem = (unsigned char *)bigphysarea_alloc_pages(n,0,GFP_KERNEL);
         if (mem==0)
         {
            printk(MV300_ERR ": bigphysarea_alloc_pages for V4L bufs failed\n");
            v4l_fbuffer_free(btv);
            return -ENOBUFS;
         }
         btv->v4l_gbuf[i].fbuffer      = mem;
         btv->v4l_gbuf[i].fbuffer_phys = virt_to_phys(mem);
         btv->v4l_gbuf[i].fbuffer_bus  = virt_to_bus (mem);
         DEBUG(printk(MV300_INFO ": Bigphysarea frame %d mem 0x%x (bus: 0x%x=%d)\n",
                      i,mem,virt_to_bus(mem),virt_to_bus(mem));)

         /* Zero out the allocated memory */
         memset(btv->v4l_gbuf[i].fbuffer,0,v4l_bufsize);
#else
#if defined(MV300_USE_HIMEM)

         /* Use high memory which has been left at boot time */

         /* Ok., Ok. this is an evil hack - we make the assumption that
            physical addresses are the same as bus addresses (true at least
            for Intel processors). The whole method of obtaining and using
            this memory is not very nice - but I hope it saves some
            poor users from kernel hacking, which might have even more
            evil results */

         if (i==0)
         {
            pmem = get_high_mem(v4l_nbufs*v4l_bufsize);
            if (pmem==0)
            {
               printk(MV300_ERR ": get_high_mem for V4L bufs failed\n");
               return -ENOBUFS;
            }
            btv->v4l_gbuf[0].fbuffer      = 0;
            btv->v4l_gbuf[0].fbuffer_phys = pmem;
            btv->v4l_gbuf[0].fbuffer_bus  = pmem;
            printk(MV300_INFO ": Using %d KB high memory\n",(v4l_nbufs*v4l_bufsize)>>10);
         }
         else
         {
            btv->v4l_gbuf[i].fbuffer      = 0;
            btv->v4l_gbuf[i].fbuffer_phys = pmem + i*v4l_bufsize;
            btv->v4l_gbuf[i].fbuffer_bus  = pmem + i*v4l_bufsize;
         }
#else
         /* No bigphysarea present, usage of high memory disabled,
            but user wants buffers of more than MAX_KMALLOC_MEM */
         printk(MV300_ERR ": No bigphysarea_patch present, usage of high memory disabled\n");
         printk(MV300_ERR ": but you want V4L buffers of size %d KB\n",v4l_bufsize>>10);
         return -ENOBUFS;
#endif
#endif
      }
   }

   return 0;
}

/* free the V4L grab buffers */
static void v4l_fbuffer_free(struct buz * btv)
{
   int i, off;
   unsigned char *mem;

   for (i = 0; i<v4l_nbufs; i++)
   {
      if(!btv->v4l_gbuf[i].fbuffer) continue;

      if (v4l_bufsize <= MAX_KMALLOC_MEM)
      {
         mem = btv->v4l_gbuf[i].fbuffer;
         for(off=0;off<v4l_bufsize;off+=PAGE_SIZE) mem_map_unreserve(MAP_NR(mem+off));
         kfree((void *)btv->v4l_gbuf[i].fbuffer);
      }
#if defined(CONFIG_BIGPHYS_AREA)
      else
         bigphysarea_free_pages((void *)btv->v4l_gbuf[i].fbuffer);
#endif
      btv->v4l_gbuf[i].fbuffer = NULL;
   }
}
/*
 *   Allocate the MJPEG grab buffers.
 *
 *   If the requested buffer size is smaller than MAX_KMALLOC_MEM,
 *   kmalloc is used to request a physically contiguous area,
 *   else we allocate the memory in framgents with get_free_page.
 *
 *   If a Natoma chipset is present and this is a revision 1 zr36057,
 *   each MJPEG buffer needs to be physically contiguous.
 *   (RJ: This statement is from Dave Perks' original driver,
 *   I could never check it because I have a zr36067)
 *   The driver cares about this because it reduces the buffer
 *   size to MAX_KMALLOC_MEM in that case (which forces contiguous allocation).
 *
 *   RJ: The contents grab buffers needs never be accessed in the driver.
 *       Therefore there is no need to allocate them with vmalloc in order
 *       to get a contiguous virtual memory space.
 *       I don't understand why many other drivers first allocate them with
 *       vmalloc (which uses internally also get_free_page, but delivers you
 *       virtual addresses) and then again have to make a lot of efforts
 *       to get the physical address.
 *
 */

static int jpg_fbuffer_alloc(struct buz * btv)
{
   int i, j, off, alloc_contig;
   unsigned long mem;

   /* Decide if we should alloc contiguous or fragmented memory */
   /* This has to be identical in jpg_fbuffer_alloc and jpg_fbuffer_free */

   alloc_contig = (btv->jpg_bufsize < MAX_KMALLOC_MEM);

   for (i = 0; i < btv->jpg_nbufs; i++)
   {
      if(btv->jpg_gbuf[i].frag_tab)
         printk(MV300_WARNING ": jpg_fbuffer_alloc: buffer %d allready allocated ???\n",i);

      /* Allocate fragment table for this buffer */

      mem = get_free_page(GFP_KERNEL);
      if (mem==0) 
      {
         printk(MV300_ERR ": jpg_fbuffer_alloc: get_free_page (frag_tab) failed for buffer %d\n",i);
         jpg_fbuffer_free(btv);
         return -ENOBUFS;
      }
      memset((void*)mem,0,PAGE_SIZE);
      btv->jpg_gbuf[i].frag_tab     = (u32*) mem;
      btv->jpg_gbuf[i].frag_tab_bus = virt_to_bus((void*)mem);

      if(alloc_contig)
      {
         mem = (unsigned long)kmalloc(btv->jpg_bufsize,GFP_KERNEL);
         if(mem==0)
         {
            printk(MV300_ERR ": jpg_fbuffer_alloc: kmalloc failed for buffer %d\n",i);
            jpg_fbuffer_free(btv);
            return -ENOBUFS;
         }
         btv->jpg_gbuf[i].frag_tab[0] = virt_to_bus((void*)mem);
         btv->jpg_gbuf[i].frag_tab[1] = ((btv->jpg_bufsize/4)<<1)|1;
         for(off=0;off<btv->jpg_bufsize;off+=PAGE_SIZE) mem_map_reserve(MAP_NR(mem+off));
      }
      else
      {
         /* jpg_bufsize is allreay page aligned */
         for (j=0; j<btv->jpg_bufsize/PAGE_SIZE; j++)
         {
            mem = get_free_page(GFP_KERNEL);
            if(mem==0)
            {
               printk(MV300_ERR ": jpg_fbuffer_alloc: get_free_page failed for buffer %d\n",i);
               jpg_fbuffer_free(btv);
               return -ENOBUFS;
            }

            btv->jpg_gbuf[i].frag_tab[2*j  ] = virt_to_bus((void*)mem);
            btv->jpg_gbuf[i].frag_tab[2*j+1] = (PAGE_SIZE/4)<<1;
            mem_map_reserve(MAP_NR(mem));
         }

         btv->jpg_gbuf[i].frag_tab[2*j-1] |= 1;
      }
   }

   DEBUG(printk("jpg_fbuffer_alloc: %d KB allocated\n",
                (btv->jpg_nbufs*btv->jpg_bufsize)>>10);)
   btv->jpg_buffers_allocated = 1;
   return 0;
}

/* free the MJPEG grab buffers */
static void jpg_fbuffer_free(struct buz * btv)
{
   int i, j, off, alloc_contig;
   unsigned char *mem;

   /* Decide if we should alloc contiguous or fragmented memory */
   /* This has to be identical in jpg_fbuffer_alloc and jpg_fbuffer_free */

   alloc_contig = (btv->jpg_bufsize < MAX_KMALLOC_MEM);

   for (i = 0; i < btv->jpg_nbufs; i++)
   {
      if(!btv->jpg_gbuf[i].frag_tab) continue;

      if(alloc_contig)
      {
         if(btv->jpg_gbuf[i].frag_tab[0])
         {
            mem = (unsigned char *) bus_to_virt(btv->jpg_gbuf[i].frag_tab[0]);
            for(off=0;off<btv->jpg_bufsize;off+=PAGE_SIZE) mem_map_unreserve(MAP_NR(mem+off));
            kfree((void *) mem);
            btv->jpg_gbuf[i].frag_tab[0] = 0;
            btv->jpg_gbuf[i].frag_tab[1] = 0;
         }
      }
      else
      {
         for(j=0; j<btv->jpg_bufsize/PAGE_SIZE; j++)
         {
            if(!btv->jpg_gbuf[i].frag_tab[2*j]) break;
            mem_map_unreserve(MAP_NR(bus_to_virt(btv->jpg_gbuf[i].frag_tab[2*j])));
            free_page((unsigned long)bus_to_virt(btv->jpg_gbuf[i].frag_tab[2*j]));
            btv->jpg_gbuf[i].frag_tab[2*j  ] = 0;
            btv->jpg_gbuf[i].frag_tab[2*j+1] = 0;
         }
      }

      free_page( (unsigned long)btv->jpg_gbuf[i].frag_tab );
      btv->jpg_gbuf[i].frag_tab = NULL;
   }
   btv->jpg_buffers_allocated = 0;
}


/* ----------------------------------------------------------------------- */

/* I2C functions                                                           */

#define I2C_DELAY   5


/* software I2C functions */

static void i2c_setlines(struct i2c_bus * bus, int ctrl, int data)
{
   struct buz * btv = (struct buz *)bus->data;
   btwrite((data << 1) | ctrl, ZR36057_I2CBR);
   udelay(I2C_DELAY);
}

static int i2c_getdataline(struct i2c_bus * bus)
{
   struct buz * btv = (struct buz *)bus->data;
   return (btread(ZR36057_I2CBR) & ZR36057_I2CBR_SDA);
}

void attach_inform(struct i2c_bus * bus, int id)
{
   DEBUG(struct buz * btv = (struct buz *)bus->data);
   DEBUG(printk(MV300_DEBUG "-%u: i2c attach %02x\n", btv->id, id));
}

void detach_inform(struct i2c_bus * bus, int id)
{
   DEBUG(struct buz * btv = (struct buz *)bus->data);
   DEBUG(printk(MV300_DEBUG "-%u: i2c detach %02x\n", btv->id, id));
}

static struct i2c_bus mv300_i2c_bus_template =
{
   "zr36057",
   I2C_BUSID_BT848,
   NULL,

   SPIN_LOCK_UNLOCKED,

   attach_inform,
   detach_inform,

   i2c_setlines,
   i2c_getdataline,
   NULL,
   NULL,
};


/* ----------------------------------------------------------------------- */

static void GPIO(struct buz * btv, unsigned bit, unsigned value)
{
   u32 reg;
   u32 mask;

   mask = 1 << (24 + bit);
   reg = btread(ZR36057_GPPGCR1) & ~mask;
   if (value) {
      reg |= mask;
   }
   btwrite(reg, ZR36057_GPPGCR1);
}


/*
 *   Set the registers for the size we have specified. Don't bother
 *   trying to understand this without the ZR36057 manual in front of
 *   you [AC].
 *
 *   PS: The manual is free for download in .pdf format from
 *   www.zoran.com - nicely done those folks.
 */

struct tvnorm
{
   u16 Wt, Wa, Ht, Ha, HStart, VStart;
};

static struct tvnorm tvnorms[] = {
   /* PAL-BDGHI */
   { 864, 720, 625, 576, 31, 16 },
   /* NTSC */
   { 858, 720, 525, 480, 66, 15 },
};
#define TVNORMS (sizeof(tvnorms) / sizeof(tvnorm))

static int format2bpp(int format)
{
   int bpp;

   /* Determine the number of bytes per pixel for the video format requested */

   switch (format) {

   case VIDEO_PALETTE_YUV422:
      bpp = 2;
      break;

   case VIDEO_PALETTE_RGB555:
      bpp = 2;
      break;

   case VIDEO_PALETTE_RGB565:
      bpp = 2;
      break;

   case VIDEO_PALETTE_RGB24:
      bpp = 3;
      break;

   case VIDEO_PALETTE_RGB32:
      bpp = 4;
      break;

   default:
      bpp = 0;
   }

   return bpp;
}

/*
 * set geometry
 */
static void zr36057_set_vfe(struct buz * btv, int video_width, int video_height,
                            unsigned int video_format)
{
   struct tvnorm * tvn;
   unsigned HStart, HEnd, VStart, VEnd;
   unsigned DispMode;
   unsigned VidWinWid, VidWinHt;
   unsigned hcrop1, hcrop2, vcrop1, vcrop2;
   unsigned Wa, We, Ha, He;
   unsigned X, Y, HorDcm, VerDcm;
   u32 reg;
   unsigned mask_line_size;

   if(btv->params.norm<0 || btv->params.norm>1) {
      printk(MV300_ERR ": set_vfe: video_norm = %d not valid\n",btv->params.norm);
      return;
   }
   if(video_width<BUZ_MIN_WIDTH || video_height<BUZ_MIN_HEIGHT) {
      printk(MV300_ERR ": set_vfe: w=%d h=%d not valid\n",video_width,video_height);
      return;
   }

   tvn = &tvnorms[btv->params.norm];

   Wa = tvn->Wa;
   Ha = tvn->Ha;

   /* if window has more than half of active height,
      switch on interlacing - we want the full information */

   btv->video_interlace = (video_height > Ha / 2);

   /**** zr36057 ****/

   /* horizontal */
   VidWinWid = video_width;
   X = (VidWinWid * 64 + tvn->Wa - 1) / tvn->Wa;
   We = (VidWinWid * 64) / X;
   HorDcm = 64 - X;
   hcrop1 = 2 * ((tvn->Wa - We) / 4);
   hcrop2 = tvn->Wa - We - hcrop1;
   HStart = tvn->HStart | 1;
   HEnd = HStart + tvn->Wa - 1;
   HStart += hcrop1;
   HEnd -= hcrop2;
   reg = ((HStart & ZR36057_VFEHCR_Hmask) << ZR36057_VFEHCR_HStart)
       | ((HEnd & ZR36057_VFEHCR_Hmask)   << ZR36057_VFEHCR_HEnd);
   reg |= ZR36057_VFEHCR_HSPol;
   btwrite(reg, ZR36057_VFEHCR);

   /* Vertical */
   DispMode = !btv->video_interlace;
   VidWinHt = DispMode ? video_height : video_height / 2;
   Y = (VidWinHt * 64 * 2 + tvn->Ha - 1) / tvn->Ha;
   He = (VidWinHt * 64) / Y;
   VerDcm = 64 - Y;
   vcrop1 = (tvn->Ha / 2 - He) / 2;
   vcrop2 = tvn->Ha / 2 - He - vcrop1;
   VStart = tvn->VStart;
   VEnd = VStart + tvn->Ha / 2 - 1;
   VStart += vcrop1;
   VEnd -= vcrop2;
   reg = ((VStart & ZR36057_VFEVCR_Vmask) << ZR36057_VFEVCR_VStart)
       | ((VEnd & ZR36057_VFEVCR_Vmask)   << ZR36057_VFEVCR_VEnd);
   reg |= ZR36057_VFEVCR_VSPol;
   btwrite(reg, ZR36057_VFEVCR);

   /* scaler and pixel format */
   reg = ZR36057_VFESPFR_ExtFl /* Trying to live with ExtFI */
      | (HorDcm << ZR36057_VFESPFR_HorDcm)
      | (VerDcm << ZR36057_VFESPFR_VerDcm)
      | (DispMode << ZR36057_VFESPFR_DispMode)
      | ZR36057_VFESPFR_LittleEndian;
   /* RJ: I don't know, why the following has to be the opposite
          of the corresponding ZR36060 setting, but only this way
          we get the correct colors when uncompressing to the screen  */
   reg |= ZR36057_VFESPFR_VCLKPol;
   /* RJ: Don't know if that is needed for NTSC also */
   reg |= ZR36057_VFESPFR_TopField;
   switch (video_format) {

   case VIDEO_PALETTE_YUV422:
      reg |= ZR36057_VFESPFR_YUV422;
      break;

   case VIDEO_PALETTE_RGB555:
      reg |= ZR36057_VFESPFR_RGB555 | ZR36057_VFESPFR_ErrDif;
      break;

   case VIDEO_PALETTE_RGB565:
      reg |= ZR36057_VFESPFR_RGB565 | ZR36057_VFESPFR_ErrDif;
      break;

   case VIDEO_PALETTE_RGB24:
      reg |= ZR36057_VFESPFR_RGB888 | ZR36057_VFESPFR_Pack24;
      break;

   case VIDEO_PALETTE_RGB32:
      reg |= ZR36057_VFESPFR_RGB888;
      break;

   default:
      printk(MV300_INFO "-%u: Unknown color_fmt=%x\n", btv->id, video_format);
      return;

   }
   if (HorDcm >= 48) {
      reg |= 3 << ZR36057_VFESPFR_HFilter; /* 5 tap filter */
   } else if (HorDcm >= 32) {
      reg |= 2 << ZR36057_VFESPFR_HFilter; /* 4 tap filter */
   } else if (HorDcm >= 16) {
      reg |= 1 << ZR36057_VFESPFR_HFilter; /* 3 tap filter */
   }
   btwrite(reg, ZR36057_VFESPFR);

   /* display configuration */

   reg = (16 << ZR36057_VDCR_MinPix)
       | (VidWinHt << ZR36057_VDCR_VidWinHt)
       | (VidWinWid << ZR36057_VDCR_VidWinWid);
   if (triton)
      reg &= ~ZR36057_VDCR_Triton;
   else
      reg |= ZR36057_VDCR_Triton;
   btwrite(reg, ZR36057_VDCR);

   /* Write overlay clipping mask data, but don't enable overlay clipping */
   /* RJ: since this makes only sense on the screen, we use 
      btv->window.width instead of video_width */

   mask_line_size = (BUZ_MAX_WIDTH + 31) / 32;
   reg = virt_to_bus(btv->overlay_mask);
   btwrite(reg, ZR36057_MMTR);
   reg = virt_to_bus(btv->overlay_mask + mask_line_size);
   btwrite(reg, ZR36057_MMBR);
   reg = mask_line_size - (btv->window.width+31)/32;
   if (DispMode == 0) reg += mask_line_size;
   reg <<= ZR36057_OCR_MaskStride;
   btwrite(reg, ZR36057_OCR);

}

/*
 * Switch overlay on or off
 */

static void zr36057_overlay(struct buz * btv, int on)
{
   int fmt, bpp;
   u32 reg;

   if (on)
   {
      /* do the necessary settings ... */

      btand(~ZR36057_VDCR_VidEn, ZR36057_VDCR); /* switch it off first */

      switch(btv->buffer.depth)
      {
         case 15: fmt = VIDEO_PALETTE_RGB555; bpp = 2; break;
         case 16: fmt = VIDEO_PALETTE_RGB565; bpp = 2; break;
         case 24: fmt = VIDEO_PALETTE_RGB24;  bpp = 3; break;
         case 32: fmt = VIDEO_PALETTE_RGB32;  bpp = 4; break;
         default: fmt = 0; bpp=0;
      }

      zr36057_set_vfe(btv,btv->window.width,btv->window.height,fmt);

      /* Start and length of each line MUST be 4-byte aligned.
         This should be allready checked before the call to this routine.
         All error messages are internal driver checking only! */

      /* video display top and bottom registers */

      reg = (u32)btv->buffer.base
          + btv->window.x * bpp
          + btv->window.y * btv->buffer.bytesperline;
      btwrite(reg, ZR36057_VDTR);
      if(reg&3) printk(MV300_ERR ": zr36057_overlay: video_address not aligned\n");
      if (btv->video_interlace) reg += btv->buffer.bytesperline;
      btwrite(reg, ZR36057_VDBR);

      /* video stride, status, and frame grab register */

      reg = btv->buffer.bytesperline - btv->window.width * bpp;
      if (btv->video_interlace) reg += btv->buffer.bytesperline;
      if(reg&3) printk(MV300_ERR ": zr36057_overlay: video_stride not aligned\n");
      reg = (reg << ZR36057_VSSFGR_DispStride);
      reg |= ZR36057_VSSFGR_VidOvf; /* clear overflow status */
      btwrite(reg, ZR36057_VSSFGR);

      /* Set overlay clipping */

      if(btv->window.clipcount) btor(ZR36057_OCR_OvlEnable,ZR36057_OCR);

      /* ... and switch it on */

      btor(ZR36057_VDCR_VidEn, ZR36057_VDCR);
   }
   else
   {
      /* Switch it off */

      btand(~ZR36057_VDCR_VidEn, ZR36057_VDCR);
   }
}

/*
 * The overlay mask has one bit for each pixel on a scan line,
 *  and the maximum window size is BUZ_MAX_WIDTH * BUZ_MAX_HEIGHT pixels.
 */
static void write_overlay_mask(struct buz * btv, struct video_clip * vp, int count)
{
   unsigned mask_line_size = (BUZ_MAX_WIDTH + 31) / 32;
   u32 * mask;
   int x, y, width, height;
   unsigned i, j, k;
   u32 reg;

   /* fill mask with one bits */
   memset(btv->overlay_mask, ~0, mask_line_size * 4 * BUZ_MAX_HEIGHT);
   reg = 0;

   for (i = 0; i < count; ++i) {
      /* pick up local copy of clip */
      x = vp[i].x;
      y = vp[i].y;
      width  = vp[i].width;
      height = vp[i].height;

      /* trim clips that extend beyond the window */
      if (x < 0) {
         width += x;
         x = 0;
      }
      if (y < 0) {
         height += y;
         y = 0;
      }
      if (x + width > btv->window.width) {
         width = btv->window.width - x;
      }
      if (y + height > btv->window.height) {
         height = btv->window.height - y;
      }

      /* ignore degenerate clips */
      if (height <= 0) {
         continue;
      }
      if (width <= 0) {
         continue;
      }

      /* apply clip for each scan line */
      for (j = 0; j < height; ++j) {
         /* reset bit for each pixel */
         /* this can be optimized later if need be */
         mask = btv->overlay_mask + (y + j) * mask_line_size;
         for (k = 0; k < width; ++k) {
            mask[(x + k) / 32] &= ~((u32)1 << (x + k) % 32);
         }
      }
   }
}

/* Enable/Disable uncompressed memory grabbing of the 36057 */

static void zr36057_set_memgrab(struct buz * btv, int mode)
{
   if (mode)
   {
      if(btread(ZR36057_VSSFGR) & (ZR36057_VSSFGR_SnapShot|ZR36057_VSSFGR_FrameGrab))
         printk(MV300_WARNING "-%u: zr36057_set_memgrab_on with SnapShot or FrameGrab on ???\n",btv->id);

      /* switch on VSync interrupts */

      btwrite(IRQ_MASK, ZR36057_ISR); // Clear Interrupts
      btor(ZR36057_ICR_GIRQ0, ZR36057_ICR);

      /* enable SnapShot */

      btor(ZR36057_VSSFGR_SnapShot, ZR36057_VSSFGR);

      /* Set zr36057 video front end  and enable video */

#ifdef XAWTV_HACK
      zr36057_set_vfe(btv, btv->gwidth>720 ? 720 : btv->gwidth, btv->gheight, btv->gformat);
#else
      zr36057_set_vfe(btv, btv->gwidth, btv->gheight, btv->gformat);
#endif

      btv->v4l_memgrab_active = 1;
   }
   else
   {
      btv->v4l_memgrab_active = 0;

      /* switch off VSync interrupts */

      btand(~ZR36057_ICR_GIRQ0, ZR36057_ICR);

      /* reenable grabbing to screen if it was running */

      if(btv->v4l_overlay_active)
      {
         zr36057_overlay(btv,1);
      }
      else
      {
         btand(~ZR36057_VDCR_VidEn, ZR36057_VDCR);
         btand(~ZR36057_VSSFGR_SnapShot, ZR36057_VSSFGR);
      }
   }
}

static int wait_grab_pending(struct buz * btv)
{
   unsigned long flags;

   /* wait until all pending grabs are finished */

   if(!btv->v4l_memgrab_active) return 0;

   while(btv->v4l_pend_tail != btv->v4l_pend_head)
   {
      interruptible_sleep_on(&btv->v4l_capq);
      if (signal_pending(current)) return -ERESTARTSYS;
   }

   spin_lock_irqsave(&btv->lock, flags);
   zr36057_set_memgrab(btv,0);
   spin_unlock_irqrestore(&btv->lock, flags);

   return 0;
}

/*
 *   V4L Buffer grabbing
 */

static int v4l_grab(struct buz * btv, struct video_mmap * mp)
{
   unsigned long flags;
   int res, bpp;

   /*
    * There is a long list of limitations to what is allowed to be grabbed
    * We don't output error messages her, since some programs (e.g. xawtv)
    * just try several settings to find out what is valid or not.
    */

   /* No grabbing outside the buffer range! */

   if (mp->frame >= v4l_nbufs || mp->frame < 0) return -EINVAL;

   /* Check size and format of the grab wanted */

   if (mp->height < BUZ_MIN_HEIGHT || mp->width < BUZ_MIN_WIDTH) return -EINVAL;
   if (mp->height > BUZ_MAX_HEIGHT || mp->width > BUZ_MAX_WIDTH) return -EINVAL;

   bpp = format2bpp(mp->format);
   if( bpp == 0 ) return -EINVAL;

   /* Check against available buffer size */

   if(mp->height*mp->width*bpp > v4l_bufsize) return -EINVAL;

   /* The video front end needs 4-byte alinged line sizes */

   if( (bpp==2 && (mp->width&1)) || (bpp==3 && (mp->width&3)) ) return -EINVAL;

   /*
    * To minimize the time spent in the IRQ routine, we avoid setting up
    * the video front end there.
    * If this grab has different parameters from a running streaming capture
    * we stop the streaming capture and start it over again.
    */

   if( btv->v4l_memgrab_active && 
       (btv->gwidth!=mp->width || btv->gheight!=mp->height || btv->gformat!=mp->format) )
   {
      res = wait_grab_pending(btv);
      if(res) return res;
   }
   btv->gwidth  = mp->width;
   btv->gheight = mp->height;
   btv->gformat = mp->format;
   btv->gbpl    = bpp*btv->gwidth;


   spin_lock_irqsave(&btv->lock, flags);

   /* make sure a grab isn't going on currently with this buffer */

   switch (btv->v4l_gbuf[mp->frame].state) {

   default:
   case BUZ_STATE_PEND:
      res = -EBUSY;         /* what are you doing? */
      break;

   case BUZ_STATE_USER:
   case BUZ_STATE_DONE:
      /* since there is at least one unused buffer there's room for at least one more pend[] entry */
      btv->v4l_pend[btv->v4l_pend_head++ & V4L_MASK_FRAME] = mp->frame;
      btv->v4l_gbuf[mp->frame].state  = BUZ_STATE_PEND;
      res = 0;
      break;

   }

   /* put the 36057 into frame grabbing mode */

   if(!res && !btv->v4l_memgrab_active) zr36057_set_memgrab(btv, 1);

   spin_unlock_irqrestore(&btv->lock, flags);

   return res;
}

/*
 * Sync on a V4L buffer
 */

static int v4l_sync(struct buz * btv, int frame)
{
   unsigned long flags;


   /* check passed-in frame number */
   if (frame >= v4l_nbufs || frame < 0) {
      printk(MV300_ERR ": v4l_sync: frame %d is invalid\n",frame);
      return -EINVAL;
   }

   /* Check if is buffer was queued at all */

   if (btv->v4l_gbuf[frame].state == BUZ_STATE_USER) {
      printk(MV300_ERR ": v4l_sync: Trying to sync on a buffer which was not queued?\n");
      return -EPROTO;
   }

   /* wait on this buffer to get ready */

   while(btv->v4l_gbuf[frame].state == BUZ_STATE_PEND)
   {
      interruptible_sleep_on(&btv->v4l_capq);
      if (signal_pending(current)) return -ERESTARTSYS;
   }

   /* buffer should now be in BUZ_STATE_DONE */

   if(btv->v4l_gbuf[frame].state != BUZ_STATE_DONE)
      printk(MV300_ERR ": v4l_sync - internal error\n");

   /* Check if streaming capture has finished */

   spin_lock_irqsave(&btv->lock, flags);

   if(btv->v4l_pend_tail==btv->v4l_pend_head) zr36057_set_memgrab(btv,0);

   spin_unlock_irqrestore(&btv->lock, flags);

   return 0;
}
/*****************************************************************************
 *                                                                           *
 *  Set up the MV300-specific MJPEG part                                     *
 *                                                                           *
 *****************************************************************************/

/*
Wait til post office is no longer busy */
static int post_office_wait(struct buz * btv)
{
   u32 por;

   while (((por = btread(ZR36057_POR)) & (ZR36057_POR_POPen | ZR36057_POR_POTime)) == ZR36057_POR_POPen) {
      /* wait for something to happen */
   }
   if ((por & ZR36057_POR_POPen) != 0) {
      printk(MV300_INFO "-%u: pop pending %08x\n", btv->id, por);
      return -1;
   }
   if ((por & (ZR36057_POR_POTime | ZR36057_POR_POPen)) != 0) {
      printk(MV300_INFO "-%u: pop timeout %08x\n", btv->id, por);
      return -1;
   }
   return 0;
}

static int post_office_write(struct buz * btv, unsigned guest, unsigned reg, unsigned value)
{
   u32 por;

   por = ZR36057_POR_PODir | ZR36057_POR_POTime | ((guest & 7) << 20) | ((reg & 7) << 16) | (value & 0xFF);
   btwrite(por, ZR36057_POR);
   return post_office_wait(btv);
}

static int post_office_read(struct buz * btv, unsigned guest, unsigned reg)
{
   u32 por;

   por = ZR36057_POR_POTime | ((guest & 7) << 20) | ((reg & 7) << 16);
   btwrite(por, ZR36057_POR);
   if (post_office_wait(btv) < 0) {
      return -1;
   }
   return btread(ZR36057_POR) & 0xFF;
}

static int zr36060_write_8(struct buz * btv, unsigned reg, unsigned val)
{
   if (post_office_wait(btv)
       ||  post_office_write(btv, 0, 1, reg >> 8)
       ||  post_office_write(btv, 0, 2, reg)) {
      return -1;
   }
   return post_office_write(btv, 0, 3, val);
}

static int zr36060_write_16(struct buz * btv, unsigned reg, unsigned val)
{
   if (zr36060_write_8(btv, reg + 0, val >> 8)) {
      return -1;
   }
   return zr36060_write_8(btv, reg + 1, val >> 0);
}

static int zr36060_write_24(struct buz * btv, unsigned reg, unsigned val)
{
   if (zr36060_write_8(btv, reg + 0, val >> 16)) {
      return -1;
   }
   return zr36060_write_16(btv, reg + 1, val >> 0);
}

static int zr36060_write_32(struct buz * btv, unsigned reg, unsigned val)
{
   if (zr36060_write_16(btv, reg + 0, val >> 16)) {
      return -1;
   }
   return zr36060_write_16(btv, reg + 2, val >> 0);
}

static u32 zr36060_read_8(struct buz * btv, unsigned reg)
{
   if (post_office_wait(btv)
       ||  post_office_write(btv, 0, 1, reg >> 8)
       ||  post_office_write(btv, 0, 2, reg)) {
      return -1;
   }
   return post_office_read(btv, 0, 3) & 0xFF;
}

static int zr36060_reset(struct buz * btv)
{
   return post_office_write(btv, 3, 0, 0);
}

static void zr36060_sleep(struct buz * btv, int sleep)
{
	GPIO(btv, 1, !sleep);
}


static void zr36060_set_jpg(struct buz * btv, enum buz_codec_mode mode)
{
   struct tvnorm *tvn;
   u32 reg;
   int size;

   reg   = (1 << 0)   /* CodeMstr */
         | (0 << 2)   /* CFIS=0 */
         | (0 << 6)   /* Endian=0 */
         | (0 << 7);  /* Code16=0 */
   zr36060_write_8(btv, 0x002, reg);

   switch (mode) {

   case BUZ_MODE_MOTION_DECOMPRESS:
   case BUZ_MODE_STILL_DECOMPRESS:
      reg   = 0x00;      /* Codec mode = decompression */
      break;

   case BUZ_MODE_MOTION_COMPRESS:
   case BUZ_MODE_STILL_COMPRESS:
   default:
	   reg   = 0xa4;      /* Codec mode = compression with variable scale factor */
	   break;

   }
   zr36060_write_8(btv, 0x003, reg);

   reg   = 0x00;      /* reserved, mbz */
   zr36060_write_8(btv, 0x004, reg);

   reg   = 0xff;      /* 510 bits/block */
   zr36060_write_8(btv, 0x005, reg);

   /* JPEG markers */
   reg = (btv->params.jpeg_markers) & 0x38; /* DRI, DQT, DHT */
   if(btv->params.COM_len) reg |= JPEG_MARKER_COM;
   if(btv->params.APP_len) reg |= JPEG_MARKER_APP;
   zr36060_write_8(btv, 0x006, reg);

   reg   = (0 << 3)   /* DATERR=0 */
         | (0 << 2)   /* END=0 */
         | (0 << 1)   /* EOI=0 */
         | (0 << 0);  /* EOAV=0 */
   zr36060_write_8(btv, 0x007, reg);

   /* code volume */

   /* Target field size in pixels: */
   tvn = &tvnorms[btv->params.norm];
   size = (tvn->Ha/2)*(tvn->Wa)/(btv->params.HorDcm)/(btv->params.VerDcm);

   /* Target compressed field size in bits: */
   size = size*16; /* uncompressed size in bits */
   size = size*btv->params.quality/200; /* quality = 100 is a compression ratio 2:1 */

   /* Lower limit (arbitrary, 1 KB) */
   if(size < 8192) size = 8192;

   /* Upper limit: 7/8 of the code buffers */
   if(size*btv->params.field_per_buff > btv->jpg_bufsize*7)
      size = btv->jpg_bufsize*7 / btv->params.field_per_buff;

   reg = size;
   zr36060_write_32(btv, 0x009, reg);

   /* how do we set initial SF as a function of quality parameter? */
   reg   = 0x0100;   /* SF=1.0 */
   zr36060_write_16(btv, 0x011, reg);

   reg   = 0x00ffffff;   /* AF=max */
   zr36060_write_24(btv, 0x013, reg);

   reg   = 0x0000;   /* test */
   zr36060_write_16(btv, 0x024, reg);
}

static void zr36060_set_video(struct buz * btv, enum buz_codec_mode mode)
{
   struct tvnorm *tvn;
   u32 reg;

   reg   = (0 << 7)   /* Video8=0 */
         | (0 << 6)   /* Range=0 */
         | (1 << 3)   /* FIDet=1 *//* No effect this value ? */
         | (1 << 2)   /* FIVedge=1 */
         | (1 << 1)   /* FIExt=1 *//* No effect this value ? */
         | (0 << 0);  /* SyncMstr=0 */

   /* According to ZR36067 documentation, FlDet should correspond
      to the odd_even flag of the ZR36067
   if(btv->params.odd_even) reg |= (1<<3); */ 

   if (mode != BUZ_MODE_STILL_DECOMPRESS) {
      /* limit pixels to range 16..235 as per CCIR-601 */
      reg |= (1 << 6);   /* Range=1 */
   }
   zr36060_write_8(btv, 0x030, reg);

   reg   = (1 << 7)   /* VCLKPol=1 */
         | (0 << 6)   /* PValPol=0 */
         | (1 << 5)   /* PoePol=1 */
         | (1 << 4)   /* SImgPol=1 */
         | (1 << 3)   /* BLPol=1 */
         | (0 << 2)   /* FIPol=0 *//* No effect this value ? */
         | (0 << 1)   /* HSPol=0 */
         | (1 << 0);  /* VSPol=1 */
   zr36060_write_8(btv, 0x031, reg);

   switch (btv->params.HorDcm)
   {
      default:
      case 1:
         reg = (0 << 0); break; /* HScale = 0 */

      case 2:
         reg = (1 << 0); break; /* HScale = 1 */

      case 4:
         reg = (2 << 0); break; /* HScale = 2 */
   }
   if(btv->params.VerDcm == 2) reg |= (1<<2);
   zr36060_write_8(btv, 0x032, reg);

   reg   = 0x80;      /* BackY */
   zr36060_write_8(btv, 0x033, reg);

   reg   = 0xe0;      /* BackU */
   zr36060_write_8(btv, 0x034, reg);

   reg   = 0xe0;      /* BackV */
   zr36060_write_8(btv, 0x035, reg);

   /* sync generator */

   tvn = &tvnorms[btv->params.norm];

   reg   = tvn->Ht - 1;           /* Vtotal */
   zr36060_write_16(btv, 0x036, reg);

   reg   = tvn->Wt - 1;           /* Htotal */
   zr36060_write_16(btv, 0x038, reg);

   reg   = 10 - 1;                /* VsyncSize */
   zr36060_write_8(btv, 0x03a, reg);

   reg   = 100 - 1;               /* HsyncSize */
   zr36060_write_8(btv, 0x03b, reg);

   reg   = tvn->VStart - 1;       /* BVstart */
   zr36060_write_8(btv, 0x03c, reg);

   reg   += tvn->Ha / 2;          /* BVend */
   zr36060_write_16(btv, 0x03e, reg);

   reg   = tvn->HStart - 1;       /* BHstart */
   zr36060_write_8(btv, 0x03d, reg);

   reg   += tvn->Wa;               /* BHend */
   zr36060_write_16(btv, 0x040, reg);

   /* active area */
   reg   = btv->params.img_y + tvn->VStart;  /* Vstart */
   zr36060_write_16(btv, 0x042, reg);

   reg  += btv->params.img_height;           /* Vend */
   zr36060_write_16(btv, 0x044, reg);

   reg   = btv->params.img_x + tvn->HStart;  /* Hstart */
   zr36060_write_16(btv, 0x046, reg);

   reg  += btv->params.img_width;            /* Hend */
   zr36060_write_16(btv, 0x048, reg);

   /* subimage area */
   reg   = btv->params.img_y + tvn->VStart;  /* SVstart */
   zr36060_write_16(btv, 0x04a, reg);

   reg  += btv->params.img_height;           /* SVend */
   zr36060_write_16(btv, 0x04c, reg);

   reg   = btv->params.img_x + tvn->HStart;  /* SHstart */
   zr36060_write_16(btv, 0x04e, reg);

   reg  +=  btv->params.img_width;           /* SHend */
   zr36060_write_16(btv, 0x050, reg);
}

static void zr36060_set_jpg_SOF(struct buz * btv)
{
   u32 reg;

   reg   = 0xffc0;   /* SOF marker */
   zr36060_write_16(btv, 0x060, reg);

   reg   = 17;      /* SOF length */
   zr36060_write_16(btv, 0x062, reg);

   reg   = 8;      /* precision 8 bits */
   zr36060_write_8(btv, 0x064, reg);

   reg   = btv->params.img_height/btv->params.VerDcm;   /* image height */
   zr36060_write_16(btv, 0x065, reg);

   reg   = btv->params.img_width /btv->params.HorDcm;   /* image width */
   zr36060_write_16(btv, 0x067, reg);

   reg   = 3;      /* 3 color components */
   zr36060_write_8(btv, 0x069, reg);

   reg   = 0x002100;   /* Y component */
   zr36060_write_24(btv, 0x06a, reg);

   reg   = 0x011101;   /* U component */
   zr36060_write_24(btv, 0x06d, reg);

   reg   = 0x021101;   /* V component */
   zr36060_write_24(btv, 0x070, reg);
}

static void zr36060_set_jpg_SOS(struct buz * btv)
{
   u32 reg;


   reg   = 0xffda;   /* SOS marker */
   zr36060_write_16(btv, 0x07a, reg);

   reg   = 12;      /* SOS length */
   zr36060_write_16(btv, 0x07c, reg);

   reg   = 3;      /* 3 color components */
   zr36060_write_8(btv, 0x07e, reg);

   reg   = 0x0000;   /* Y component */
   zr36060_write_16(btv, 0x07f, reg);

   reg   = 0x0111;   /* U component */
   zr36060_write_16(btv, 0x081, reg);

   reg   = 0x0211;   /* V component */
   zr36060_write_16(btv, 0x083, reg);

   reg   = 0x003f00;   /* Start, end spectral scans */
   zr36060_write_24(btv, 0x085, reg);
}

static void zr36060_set_jpg_DRI(struct buz * btv)
{
   u32 reg;


   reg   = 0xffdd;   /* DRI marker */
   zr36060_write_16(btv, 0x0c0, reg);

   reg   = 4;      /* DRI length */
   zr36060_write_16(btv, 0x0c2, reg);

   reg   = 8;      /* length in MCUs */
   zr36060_write_16(btv, 0x0c4, reg);
}

static void zr36060_set_jpg_DQT(struct buz * btv)
{
   unsigned i;
   unsigned adr;
   static const u8 dqt[] = {
      0xff, 0xdb,   /* DHT marker */
      0x00, 0x84,   /* DHT length */
      0x00,      /* table ID 0 */
      0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
      0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
      0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
      0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
      0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
      0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
      0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
      0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
      0x01,      /* table ID 1 */
      0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
      0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
   };

   /* write fixed quantitization tables */
   adr = 0x0cc;
   for (i = 0; i < sizeof(dqt); ++i) {
      zr36060_write_8(btv, adr++, dqt[i]);
   }
}

static void zr36060_set_jpg_DHT(struct buz * btv)
{
   unsigned i;
   unsigned adr;
   static const u8 dht[] = {
      0xff, 0xc4,   /* DHT marker */
      0x01, 0xa2,   /* DHT length */
      0x00,      /* table class 0, ID 0 */
      0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, /* # codes of length 1..8 */
      0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* # codes of length 8..16 */
      0x00,                  /* values for codes of length 2 */
      0x01, 0x02, 0x03, 0x04, 0x05,         /* values for codes of length 3 */
      0x06,                  /* values for codes of length 4 */
      0x07,                  /* values for codes of length 5 */
      0x08,                  /* values for codes of length 6 */
      0x09,                  /* values for codes of length 7 */
      0x0a,                  /* values for codes of length 8 */
      0x0b,                  /* values for codes of length 9 */
      0x01,      /* table class 0, ID 1 */
      0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,   /* # codes of length 1..8 */
      0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,   /* # codes of length 9..16 */
      0x00, 0x01, 0x02,            /* values for codes of length 2 */
      0x03,                  /* values for codes of length 3 */
      0x04,                  /* values for codes of length 4 */
      0x05,                  /* values for codes of length 5 */
      0x06,                  /* values for codes of length 6 */
      0x07,                  /* values for codes of length 7 */
      0x08,                  /* values for codes of length 8 */
      0x09,                  /* values for codes of length 9 */
      0x0a,                  /* values for codes of length 10 */
      0x0b,                  /* values for codes of length 11 */
      0x10,
      0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03,
      0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
      0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
	  0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04,
      0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04, 0x00,
      0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11,
      0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51,
      0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08,
      0x14, 0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23,
      0x33, 0x52, 0xf0, 0x15, 0x62, 0x72, 0xd1, 0x0a,
      0x16, 0x24, 0x34, 0xe1, 0x25, 0xf1, 0x17, 0x18,
      0x19, 0x1a, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x35,
      0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
      0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55,
      0x56, 0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65,
      0x66, 0x67, 0x68, 0x69, 0x6a, 0x73, 0x74, 0x75,
      0x76, 0x77, 0x78, 0x79, 0x7a, 0x82, 0x83, 0x84,
      0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93,
      0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2,
      0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
      0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
      0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8,
      0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
      0xd8, 0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6,
      0xe7, 0xe8, 0xe9, 0xea, 0xf2, 0xf3, 0xf4, 0xf5,
      0xf6, 0xf7, 0xf8, 0xf9, 0xfa
   };

   /* write fixed Huffman tables */
   adr = 0x1d4;
   for (i = 0; i < sizeof(dht); ++i) {
      zr36060_write_8(btv, adr++, dht[i]);
   }
}

static void zr36060_set_jpg_APP(struct buz * btv)
{
   unsigned adr;
   int len, i;
   u32 reg;


   len = btv->params.APP_len;
   if( len <  0 ) len =  0;
   if( len > 60 ) len = 60;

   i = btv->params.APPn;
   if(i< 0) i= 0;
   if(i>15) i=15;

   reg   = 0xffe0 + i;   /* APPn marker */
   zr36060_write_16(btv,0x380,reg);

   reg = len+2;      /* APPn len */
   zr36060_write_16(btv,0x382,reg);

   /* write APPn data */
   adr = 0x384;
   for (i = 0; i < 60; i++) {
         zr36060_write_8(btv, adr++, (i<len ? btv->params.APP_data[i] : 0) );
   }
}

static void zr36060_set_jpg_COM(struct buz * btv)
{
   unsigned adr;
   int len, i;
   u32 reg;


   len = btv->params.COM_len;
   if( len <  0 ) len =  0;
   if( len > 60 ) len = 60;

   reg   = 0xfffe;   /* COM marker */
   zr36060_write_16(btv,0x3c0,reg);

   reg = len+2;      /* COM len */
   zr36060_write_16(btv,0x3c2,reg);

   /* write COM data */
   adr = 0x3c4;
   for (i = 0; i < 60; i++) {
         zr36060_write_8(btv, adr++, (i<len ? btv->params.COM_data[i] : 0) );
   }
}

static void zr36060_set_cap(struct buz * btv, enum buz_codec_mode mode)
{
   unsigned i;
   u32 reg;

   zr36060_reset(btv);
   mdelay(10);

   reg   = (0 << 7)    /* Load=0 */
         | (1 << 0);   /* SynRst=1 */
   zr36060_write_8(btv, 0x000, reg);

   zr36060_set_jpg(btv, mode);
   zr36060_set_video(btv, mode);
   zr36060_set_jpg_SOF(btv);
   zr36060_set_jpg_SOS(btv);
   zr36060_set_jpg_DRI(btv);
   zr36060_set_jpg_DQT(btv);
   zr36060_set_jpg_DHT(btv);
   zr36060_set_jpg_APP(btv);
   zr36060_set_jpg_COM(btv);

   reg   = (1 << 7)   /* Load=1 */
         | (0 << 0);  /* SynRst=0 */
   zr36060_write_8(btv, 0x000, reg);

   /* wait for codec to unbusy */
   for (i = 0; i < 1000; ++i) {
      reg = zr36060_read_8(btv, 0x001);
      if ((reg & (1 << 7)) == 0) {
         DEBUG(printk(KERN_DEBUG "060: loaded, loops=%u\n", i));
         return;
      }
      udelay(1000);
   }
   printk(KERN_INFO "060: stuck busy, statux=%02x\n", reg);
}

static void zr36057_set_jpg(struct buz * btv, enum buz_codec_mode mode)
{
   struct tvnorm * tvn;
   u32 reg;
   int i;

   tvn = &tvnorms[btv->params.norm];

   /* assert P_Reset */
   btwrite(0, ZR36057_JPC);

   /* re-initialize DMA ring stuff */
   btv->jpg_que_head = 0;
   btv->jpg_dma_head = 0;
   btv->jpg_dma_tail = 0;
   btv->jpg_que_tail = 0;
   btv->jpg_seq_num  = 0;
   for (i = 0; i < BUZ_NUM_STAT_COM; ++i) {
      btv->stat_com[i] = 1;      /* mark as unavailable to zr36057 */
   }
   for (i = 0; i < btv->jpg_nbufs; i++) {
      btv->jpg_gbuf[i].state = BUZ_STATE_USER;   /* nothing going on */
   }

   /* MJPEG compression mode */
   switch (mode) {

   case BUZ_MODE_MOTION_COMPRESS:
   default:
      reg = ZR36057_JMC_MJPGCmpMode;
      break;

   case BUZ_MODE_MOTION_DECOMPRESS:
      reg = ZR36057_JMC_MJPGExpMode;
      reg |= ZR36057_JMC_SyncMstr;
      /* RJ: The following is experimental - improves the output to screen */
      if(btv->params.VFIFO_FB) reg |= ZR36057_JMC_VFIFO_FB;
      break;

   case BUZ_MODE_STILL_COMPRESS:
      reg = ZR36057_JMC_JPGCmpMode;
      break;

   case BUZ_MODE_STILL_DECOMPRESS:
      reg = ZR36057_JMC_JPGExpMode;
      break;

   }
   reg   |= ZR36057_JMC_JPG;
   if(btv->params.field_per_buff == 1) reg |= ZR36057_JMC_Fld_per_buff;
   btwrite(reg, ZR36057_JMC);

   /* vertical */
   btor(ZR36057_VFEVCR_VSPol, ZR36057_VFEVCR);
   reg = (6 << ZR36057_VSP_VsyncSize) | (tvn->Ht << ZR36057_VSP_FrmTot);
   btwrite(reg, ZR36057_VSP);
   reg = ( (btv->params.img_y+tvn->VStart) << ZR36057_FVAP_NAY)
       | (btv->params.img_height << ZR36057_FVAP_PAY);
   btwrite(reg, ZR36057_FVAP);

   /* horizontal */
   btor(ZR36057_VFEHCR_HSPol,ZR36057_VFEHCR);
   reg = ( (tvn->Wt-100) << ZR36057_HSP_HsyncStart) | (tvn->Wt << ZR36057_HSP_LineTot);
   btwrite(reg, ZR36057_HSP);
   reg = ( (btv->params.img_x+tvn->HStart) << ZR36057_FHAP_NAX)
       | (btv->params.img_width << ZR36057_FHAP_PAX);
   btwrite(reg, ZR36057_FHAP);

   /* field process parameters */
   if(btv->params.odd_even){
	   /* reg = 0; */
	   reg = ZR36057_FPP_Odd_Even;
   }else{
	   reg = 0;
	   /* reg = ZR36057_FPP_Odd_Even; */
   }
   btwrite(reg, ZR36057_FPP);

   /* Set proper VCLK Polarity, else colors will be wrong during playback */
   btor(ZR36057_VFESPFR_VCLKPol, ZR36057_VFESPFR);

   /* code base address and FIFO threshold */
   reg = virt_to_bus(btv->stat_com);
   btwrite(reg, ZR36057_JCBA);
   reg = 0x50;
   btwrite(reg, ZR36057_JCFT);

   /* JPEG codec guest ID */
   reg = (1 << ZR36057_JCGI_JPEGuestID) | (0 << ZR36057_JCGI_JPEGuestReg);
   btwrite(reg, ZR36057_JCGI);

   /* Code transfer guest ID */
   reg = (0 << ZR36057_MCTCR_CodGuestID) | (3 << ZR36057_MCTCR_CodGuestReg);
   reg |= ZR36057_MCTCR_CFlush;
   btwrite(reg, ZR36057_MCTCR);

   /* deassert P_Reset */
   btwrite(ZR36057_JPC_P_Reset, ZR36057_JPC);
}

static void zr36057_enable_jpg(struct buz * btv, enum buz_codec_mode mode)
{
   static int zero = 0;
   static int one = 1;

   switch (mode) {

   case BUZ_MODE_MOTION_COMPRESS:
      zr36060_set_cap(btv, mode);
      zr36057_set_jpg(btv, mode);
      i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_ENABLE_OUTPUT, &one);
      i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_INPUT, &zero);

      /* deassert P_Reset, assert Code transfer enable */
      btwrite(IRQ_MASK, ZR36057_ISR);
      btand(~ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
      break;

   case BUZ_MODE_MOTION_DECOMPRESS:
      i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_ENABLE_OUTPUT, &zero);
      i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_INPUT, &one);
      zr36060_set_cap(btv, mode);
      zr36057_set_jpg(btv, mode);

      /* deassert P_Reset, assert Code transfer enable */
      btwrite(IRQ_MASK, ZR36057_ISR);
      btand(~ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
      break;

   case BUZ_MODE_IDLE:
   default:
      /* shut down processing */
      btor(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
      btwrite(ZR36057_JPC_P_Reset, ZR36057_JPC);
      btand(~ZR36057_JMC_VFIFO_FB,ZR36057_JMC);
      btand(~ZR36057_JMC_SyncMstr,ZR36057_JMC);
      btand(~ZR36057_JMC_Go_en, ZR36057_JMC);
      btwrite(0, ZR36057_ISR);
      zr36060_reset(btv);
      i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_ENABLE_OUTPUT, &one);
      i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_INPUT, &zero);
      break;

   }
   btv->codec_mode = mode;
}

/*
 *   Queue a MJPEG buffer for capture/playback
 */

static int jpg_qbuf(struct buz * btv, int frame, enum buz_codec_mode mode)
{
   unsigned long flags;
   int res;

   /* Check if buffers are allocated */

   if(!btv->jpg_buffers_allocated)
   {
      printk(MV300_ERR ": jpg_qbuf: buffers not yet allocated\n");
      return -ENOMEM;
   }

   /* Does the user want to stop streaming? */

   if (frame < 0) {
      if (btv->codec_mode==mode) {
         zr36057_enable_jpg(btv, BUZ_MODE_IDLE);
         return 0;
      }
      else
      {
         printk(MV300_ERR  ": jpg_qbuf - stop streaming but not in streaming mode\n");
         return -EINVAL;
      }
   }

   /* No grabbing outside the buffer range! */

   if (frame >= btv->jpg_nbufs) {
      printk(MV300_ERR ": jpg_qbuf: buffer %d out of range\n",frame);
      return -EINVAL;
   }

   /* what is the codec mode right now? */

   if ( btv->codec_mode == BUZ_MODE_IDLE )
   {
      /* Ok load up the zr36060 and go */
      zr36057_enable_jpg(btv, mode);
   }
   else if ( btv->codec_mode != mode )
   {
      /* wrong codec mode active - invalid */
      printk(MV300_ERR ": jpg_qbuf - codec in wrong mode\n");
      return -EINVAL;
   }

   spin_lock_irqsave(&btv->lock, flags);

   /* make sure a grab isn't going on currently with this buffer */

   switch (btv->jpg_gbuf[frame].state) {

   default:
   case BUZ_STATE_DMA:
   case BUZ_STATE_PEND:
   case BUZ_STATE_DONE:
      res = -EBUSY;         /* what are you doing? */
      break;

   case BUZ_STATE_USER:
      /* since there is at least one unused buffer there's room for at least one more pend[] entry */
      btv->jpg_pend[btv->jpg_que_head++ & BUZ_MASK_FRAME] = frame;
      btv->jpg_gbuf[frame].state = BUZ_STATE_PEND;
      mv300_feed_stat_com(btv);
      res = 0;
      break;

   }

   spin_unlock_irqrestore(&btv->lock, flags);

   /* Start the zr36060 when the first frame is queued  */
   if(btv->jpg_que_head==1){
	   
	   /* wait ODD field
	   GPIO(btv, 3, 1); */
	   
	   btor(ZR36057_JMC_Go_en, ZR36057_JMC);
	   btwrite(ZR36057_JPC_P_Reset | ZR36057_JPC_CodTrnsEn | ZR36057_JPC_Active, ZR36057_JPC);

	   /* go capture both field
	   mdelay(10);
	   GPIO(btv, 3, 1); */
   }

   return res;
}

/*
 *   Sync on a MJPEG buffer
 */

static int jpg_sync(struct buz * btv, struct buz_sync * bs)
{
   unsigned long flags;
   int frame;

   if( btv->codec_mode != BUZ_MODE_MOTION_DECOMPRESS &&
       btv->codec_mode != BUZ_MODE_MOTION_COMPRESS ) {
      printk(MV300_ERR ": BUZIOCSYNC: - codec not in streaming mode\n");
      return -EINVAL;
   }

   while(btv->jpg_que_tail == btv->jpg_dma_tail) {
      interruptible_sleep_on(&btv->jpg_capq);
      if (signal_pending(current)) return -ERESTARTSYS;
   }

   spin_lock_irqsave(&btv->lock, flags);

   frame = btv->jpg_pend[btv->jpg_que_tail++ & BUZ_MASK_FRAME];

   /* buffer should now be in BUZ_STATE_DONE */

   if(btv->jpg_gbuf[frame].state != BUZ_STATE_DONE)
      printk(MV300_ERR ": jpg_sync - internal error\n");

   *bs = btv->jpg_gbuf[frame].bs;
   btv->jpg_gbuf[frame].state = BUZ_STATE_USER;

   spin_unlock_irqrestore(&btv->lock, flags);

   return 0;
}

/* when this is called the spinlock must be held */
static void mv300_feed_stat_com(struct buz * btv)
{
   /* move frames from pending queue to DMA */

   int frame, i, max_stat_com;

   max_stat_com = (btv->params.TmpDcm == 1) ? BUZ_NUM_STAT_COM : (BUZ_NUM_STAT_COM>>1);

   while ((btv->jpg_dma_head - btv->jpg_dma_tail) < max_stat_com
          && btv->jpg_dma_head != btv->jpg_que_head) {

      frame = btv->jpg_pend[btv->jpg_dma_head & BUZ_MASK_FRAME];
      if(btv->params.TmpDcm == 1)
      {
         /* fill 1 stat_com entry */
         i = btv->jpg_dma_head & BUZ_MASK_STAT_COM;
         btv->stat_com[i] = btv->jpg_gbuf[frame].frag_tab_bus;
      }
      else
      {
         /* fill 2 stat_com entries */
         i = (btv->jpg_dma_head & 1)*2;
         btv->stat_com[i  ] = btv->jpg_gbuf[frame].frag_tab_bus;
         btv->stat_com[i+1] = btv->jpg_gbuf[frame].frag_tab_bus;
      }
      btv->jpg_gbuf[frame].state = BUZ_STATE_DMA;
      btv->jpg_dma_head++;

   }
}

/* when this is called the spinlock must be held */
static void mv300_reap_stat_com(struct buz * btv)
{
   /* move frames from DMA queue to done queue */

   int i;
   u32 stat_com;
   unsigned int seq;
   unsigned int dif;
   int frame;
   struct buz_gbuffer * gbuf;

   /* In motion decompress we don't have a hardware frame counter,
      we just count the interrupts here */

   if (btv->codec_mode == BUZ_MODE_MOTION_DECOMPRESS) btv->jpg_seq_num++;

   while (btv->jpg_dma_tail != btv->jpg_dma_head)
   {
      if(btv->params.TmpDcm == 1)
         i = btv->jpg_dma_tail & BUZ_MASK_STAT_COM;
      else
         i = (btv->jpg_dma_tail & 1)*2+1;

      stat_com = btv->stat_com[i];

      if ((stat_com & 1) == 0) {
         return;
      }

      frame = btv->jpg_pend[btv->jpg_dma_tail & BUZ_MASK_FRAME];
      gbuf = &btv->jpg_gbuf[frame];
      get_fast_time(&gbuf->bs.timestamp);

      if (btv->codec_mode == BUZ_MODE_MOTION_COMPRESS)
      {
         gbuf->bs.length = (stat_com & 0x7fffff) >> 1;

         /* update sequence number with the help of the counter in stat_com */

         seq = stat_com >> 24;
         dif = (seq - btv->jpg_seq_num) & 0xff;
         btv->jpg_seq_num += dif;
      }
      else
      {
         gbuf->bs.length = 0;
      }
      gbuf->bs.seq = btv->params.TmpDcm==2 ? (btv->jpg_seq_num>>1) : btv->jpg_seq_num;
      gbuf->state  = BUZ_STATE_DONE;

      btv->jpg_dma_tail++;
   }
}

static void mv300_irq(int irq, void * dev_id, struct pt_regs * regs)
{
   u32 stat, astat;
   int count;
   struct buz * btv;
   unsigned long flags;

   btv = (struct buz *)dev_id;
   count = 0;

   spin_lock_irqsave(&btv->lock, flags);
   while (1) {
      /* get/clear interrupt status bits */
      stat = btread(ZR36057_ISR);
      astat = stat & IRQ_MASK;
      if (!astat) {
         break;
      }
      btwrite(astat, ZR36057_ISR);
      IDEBUG(printk(MV300_DEBUG "-%u: astat %08x stat %08x\n", btv->id, astat, stat));

#if (IRQ_MASK & ZR36057_ISR_GIRQ0)
      if (astat & ZR36057_ISR_GIRQ0) {

         /* Interrupts may still happen when btv->v4l_memgrab_active is switched off.
            We simply ignore them */

         if (btv->v4l_memgrab_active) {

/* A lot more checks should be here ... */
if ( (btread(ZR36057_VSSFGR)&ZR36057_VSSFGR_SnapShot) == 0 )
               printk(MV300_WARNING "-%u: Mv300IRQ with SnapShot off ???\n",btv->id);

            if ( btv->v4l_grab_frame != NO_GRAB_ACTIVE )
            {
               /* There is a grab on a frame going on, check if it has finished */

               if( (btread(ZR36057_VSSFGR)&ZR36057_VSSFGR_FrameGrab) == 0 )
               {
                  /* it is finished, notify the user */

                  btv->v4l_gbuf[btv->v4l_grab_frame].state  = BUZ_STATE_DONE;
                  btv->v4l_grab_frame = NO_GRAB_ACTIVE;
                  btv->v4l_grab_seq++;
                  btv->v4l_pend_tail++;
               }
            }

            if ( btv->v4l_grab_frame == NO_GRAB_ACTIVE )
               wake_up_interruptible(&btv->v4l_capq);

            /* Check if there is another grab queued */

            if ( btv->v4l_grab_frame == NO_GRAB_ACTIVE &&
                 btv->v4l_pend_tail != btv->v4l_pend_head )
            {

               int frame = btv->v4l_pend[btv->v4l_pend_tail & V4L_MASK_FRAME];
               u32 reg;

               btv->v4l_grab_frame = frame;

               /* Set zr36057 video front end and enable video */

               /* Buffer address */

               reg = btv->v4l_gbuf[frame].fbuffer_bus;
               btwrite(reg, ZR36057_VDTR);
               if (btv->video_interlace) reg += btv->gbpl;
               btwrite(reg, ZR36057_VDBR);

               /* video stride, status, and frame grab register */

#ifdef XAWTV_HACK
               reg  = (btv->gwidth>720) ? ((btv->gwidth&~3)-720)*btv->gbpl/btv->gwidth : 0;
#else
               reg  = 0;
#endif
               if (btv->video_interlace) reg += btv->gbpl;
               reg  = (reg << ZR36057_VSSFGR_DispStride);
               reg |= ZR36057_VSSFGR_VidOvf;
               reg |= ZR36057_VSSFGR_SnapShot;
               reg |= ZR36057_VSSFGR_FrameGrab;
               btwrite(reg, ZR36057_VSSFGR);

               btor(ZR36057_VDCR_VidEn, ZR36057_VDCR);
            }
         }
      }
#endif /* (IRQ_MASK & ZR36057_ISR_GIRQ0) */

#if (IRQ_MASK & ZR36057_ISR_GIRQ1)
      if (astat & ZR36057_ISR_GIRQ1) {
         unsigned csr = zr36060_read_8(btv, 0x001);
         unsigned isr = zr36060_read_8(btv, 0x008);

         IDEBUG(printk(MV300_DEBUG "-%u: ZR36057_ISR_GIRQ1 60_code=%02x 60_intr=%02x\n",
                  btv->id, csr, isr));

         btand(~ZR36057_ICR_GIRQ1, ZR36057_ICR);
         mv300_reap_stat_com(btv);
         mv300_feed_stat_com(btv);
      }
#endif /* (IRQ_MASK & ZR36057_ISR_GIRQ1) */

#if (IRQ_MASK & ZR36057_ISR_CodRepIRQ)
      if (astat & ZR36057_ISR_CodRepIRQ) {
         IDEBUG(printk(MV300_DEBUG "-%u: ZR36057_ISR_CodRepIRQ\n", btv->id));
         btand(~ZR36057_ICR_CodRepIRQ, ZR36057_ICR);
      }
#endif /* (IRQ_MASK & ZR36057_ISR_CodRepIRQ) */

#if (IRQ_MASK & ZR36057_ISR_JPEGRepIRQ)
      if ( (astat & ZR36057_ISR_JPEGRepIRQ) &&
           (btv->codec_mode == BUZ_MODE_MOTION_DECOMPRESS ||
            btv->codec_mode == BUZ_MODE_MOTION_COMPRESS) ) {
         mv300_reap_stat_com(btv);
         mv300_feed_stat_com(btv);
         wake_up_interruptible(&btv->jpg_capq);
      }
#endif /* (IRQ_MASK & ZR36057_ISR_JPEGRepIRQ) */

      count++;
      if (count > 10) {
         printk (MV300_WARNING "-%u: irq loop %d\n", btv->id, count);
         if (count > 20) {
            btwrite(0, ZR36057_ICR);
            printk(MV300_ERR "-%u: IRQ lockup, cleared int mask\n", btv->id);
            break;
         }
      }
   }
   spin_unlock_irqrestore(&btv->lock, flags);
}

/* Check a mv300_params struct for correctness, insert default params */

static int mv300_check_params(struct buz_params * params)
{
   int err = 0, err0 = 0;

   /* insert constant params */

   params->major_version = MAJOR_VERSION;
   params->minor_version = MINOR_VERSION;

   /* Check input and norm */

   if(params->input != 0 && params->input != 1)
   {
      printk(MV300_ERR " SET PARAMS: input = %d, must be 0 or 1\n",params->input);
      err++;
   }

   if(params->norm != VIDEO_MODE_PAL && params->norm != VIDEO_MODE_NTSC)
   {
      printk(MV300_ERR " SET PARAMS: norm = %d, must be 0 (PAL) or 1 (NTSC)\n",params->norm);
      err++;
   }

   /* Check decimation, set default values for decimation = 1, 2, 4 */

   switch (params->decimation)
   {
      case 1:

         params->HorDcm         = 1;
         params->VerDcm         = 1;
         params->TmpDcm         = 1;
         params->field_per_buff = 2;

         params->img_x          = 0;
         params->img_y          = 0;
         params->img_width      = 720;
         params->img_height     = tvnorms[params->norm].Ha/2;
         break;

      case 2:

         params->HorDcm         = 2;
         params->VerDcm         = 1;
         params->TmpDcm         = 2;
         params->field_per_buff = 1;

         params->img_x          = 8;
         params->img_y          = 0;
         params->img_width      = 704;
         params->img_height     = tvnorms[params->norm].Ha/2;
         break;

      case 4:

         params->HorDcm         = 4;
         params->VerDcm         = 2;
         params->TmpDcm         = 2;
         params->field_per_buff = 1;

         params->img_x          = 8;
         params->img_y          = 0;
         params->img_width      = 704;
         params->img_height     = tvnorms[params->norm].Ha/2;
         break;

      case 0:

         /* We have to check the data the user has set */

         if(params->HorDcm!=1 && params->HorDcm!=2 && params->HorDcm!=4) err0++;
         if(params->VerDcm!=1 && params->VerDcm!=2) err0++;
         if(params->TmpDcm!=1 && params->TmpDcm!=2) err0++;
         if(params->field_per_buff!=1 && params->field_per_buff!=2) err0++;

         if(params->img_x     <0) err0++;
         if(params->img_y     <0) err0++;
         if(params->img_width <0) err0++;
         if(params->img_height<0) err0++;
         if(params->img_x + params->img_width  > 720) err0++;
         if(params->img_y + params->img_height > tvnorms[params->norm].Ha/2) err0++;
         if(params->img_width  % (16*params->HorDcm) != 0 ) err0++;
         if(params->img_height % ( 8*params->VerDcm) != 0 ) err0++;

         if(err0)
         {
            printk(MV300_ERR " SET PARAMS: error in params for decimation = 0\n");
            err++;
         }
         break;

      default:
         printk(MV300_ERR " SET PARAMS: decimation = %d, must be 0, 1, 2 or 4\n",params->decimation);
         err++;
         break;
   }

   if(params->quality > 100) params->quality = 100;
   if(params->quality <   5) params->quality =   5;

   if(params->APPn <  0) params->APPn =  0;
   if(params->APPn > 15) params->APPn = 15;
   if(params->APP_len <  0) params->APP_len =  0;
   if(params->APP_len > 60) params->APP_len = 60;
   if(params->COM_len <  0) params->COM_len =  0;
   if(params->COM_len > 60) params->COM_len = 60;

   if(err) return -EINVAL;

   return 0;

}
static void mv300_open_init_params(struct buz * btv)
{
   int i;

   /* Per default, map the V4L Buffers */

   btv->map_mjpeg_buffers= 0;

   /* User must explicitly set a window */

   btv->window_set       = 0;

   btv->window.x         = 0;
   btv->window.y         = 0;
   btv->window.width     = 0;
   btv->window.height    = 0;
   btv->window.chromakey = 0;
   btv->window.flags     = 0;
   btv->window.clips     = NULL;
   btv->window.clipcount = 0;

   btv->video_interlace  = 0;

   btv->v4l_capq         = NULL;

   btv->v4l_memgrab_active = 0;
   btv->v4l_overlay_active = 0;

   btv->v4l_grab_frame    = NO_GRAB_ACTIVE;
   btv->v4l_grab_seq      = 0;

   btv->gwidth  = 0;
   btv->gheight = 0;
   btv->gformat = 0;
   btv->gbpl    = 0;

   /* DMA ring stuff for V4L */

   btv->v4l_pend_tail  = 0;
   btv->v4l_pend_head  = 0;
   for (i = 0; i < v4l_nbufs; i++) {
      btv->v4l_gbuf[i].state = BUZ_STATE_USER;   /* nothing going on */
   }

   /* Set necessary params and call mv300_check_params to set the defaults */

   btv->params.decimation = 1;

   btv->params.quality    = 50; /* default compression factor 8 */
   btv->params.odd_even   = 1;

   btv->params.APPn    = 0;
   btv->params.APP_len = 0; /* No APPn marker */
   for(i=0;i<60;i++) btv->params.APP_data[i] = 0;

   btv->params.COM_len = 0; /* No COM marker */
   for(i=0;i<60;i++) btv->params.COM_data[i] = 0;

   btv->params.VFIFO_FB = 0;

   memset(btv->params.reserved,0,sizeof(btv->params.reserved));

   btv->params.jpeg_markers = JPEG_MARKER_DHT | JPEG_MARKER_DQT;

   i = mv300_check_params(&btv->params);
   if(i) printk(MV300_ERR "mv300_open_init_params internal error\n");
}

/*
 *   Open a mv300 card. Right now the flags stuff is just playing
 */

static int mv300_open(struct video_device * dev, int flags)
{
   struct buz * btv = (struct buz *)dev;

   DEBUG(printk(MV300_INFO ": mv300_open\n");)

   switch (flags) {

   case 0:
      if (btv->user) {
         printk(MV300_WARNING ": mv300_open: MV-300 is allready in use\n");
         return -EBUSY;
      }
      btv->user++;

      if (v4l_fbuffer_alloc(btv) < 0) {
         btv->user--;
         printk(MV300_ERR ": mv300_open: v4l_fbuffer_alloc failed\n");
         return -ENOMEM;
      }

      /* default setup */

      mv300_open_init_params(btv);

      zr36057_enable_jpg(btv, BUZ_MODE_IDLE);

      btwrite(IRQ_MASK, ZR36057_ISR); // Clears interrupts
      btor(ZR36057_ICR_IntPinEn, ZR36057_ICR);

      break;

   default:
      printk(MV300_WARNING ": mv300_open: flags = 0x%x not yet supported\n",flags);
      return -EBUSY;
      break;

   }
   MOD_INC_USE_COUNT;
   return 0;
}

static void mv300_close(struct video_device * dev)
{
   struct buz * btv = (struct buz *)dev;

   DEBUG(printk(MV300_INFO ": mv300_close\n");)

   /* disable interrupts */
   btand(~ZR36057_ICR_IntPinEn, ZR36057_ICR);

   /* wake up sleeping beauties */
   wake_up_interruptible(&btv->v4l_capq);
   wake_up_interruptible(&btv->jpg_capq);

   zr36057_enable_jpg(btv, BUZ_MODE_IDLE);
   zr36057_set_memgrab(btv, 0);
   if(btv->v4l_overlay_active) zr36057_overlay(btv, 0);

   btv->user--;

   v4l_fbuffer_free(btv);
   jpg_fbuffer_free(btv);
   btv->jpg_nbufs = 0;

   MOD_DEC_USE_COUNT;
   DEBUG(printk(MV300_INFO ": mv300_close done\n");)
}


static long mv300_read(struct video_device * dev, char * buf, unsigned long count, int nonblock)
{
   return -EINVAL;
}

static long mv300_write(struct video_device * dev, const char * buf, unsigned long count, int nonblock)
{
   return -EINVAL;
}

/*
 *   ioctl routine
 */


static int mv300_ioctl(struct video_device * dev, unsigned int cmd, void * arg)
{
   struct buz * btv = (struct buz *)dev;

   switch (cmd) {

   case VIDIOCGCAP:
      {
         struct video_capability b;
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGCAP\n");)

         strncpy(b.name, btv->video_dev.name,sizeof(b.name));
         b.type = VID_TYPE_CAPTURE    |
                  VID_TYPE_OVERLAY    |
                  VID_TYPE_CLIPPING   |
                  VID_TYPE_FRAMERAM   |
                  VID_TYPE_SCALES;
         /* theoretically we could also flag VID_TYPE_SUBCAPTURE
            but this is not even implemented in the BTTV driver */

         b.channels  = 2; /* composite, svhs */
         b.audios    = 0;
         b.maxwidth  = BUZ_MAX_WIDTH;
         b.maxheight = BUZ_MAX_HEIGHT;
         b.minwidth  = BUZ_MIN_WIDTH;
         b.minheight = BUZ_MIN_HEIGHT;
         if (copy_to_user(arg, &b, sizeof(b))) {
            return -EFAULT;
         }
         return 0;
      }

   case VIDIOCGCHAN:
      {
         struct video_channel v;

         if (copy_from_user(&v, arg, sizeof(v))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGCHAN for channel %d\n",v.channel);)
         switch(v.channel) {
         case 0:
            strcpy(v.name, "Composite");
            break;
         case 1:
            strcpy(v.name, "SVHS");
            break;
         default:
            printk(MV300_ERR ": VIDIOCGCHAN on not existing channel %d\n",v.channel);
            return -EINVAL;
         }
         v.tuners = 0;
         v.flags  = 0;
         v.type   = VIDEO_TYPE_CAMERA;
         v.norm   = btv->params.norm;
         if (copy_to_user(arg, &v, sizeof(v))) {
            return -EFAULT;
         }
         return 0;
      }

   /* RJ: the documentation at http://roadrunner.swansea.linux.org.uk/v4lapi.shtml says:
    *
    * "The VIDIOCSCHAN ioctl takes an integer argument and switches the capture to this input."
    *                                 ^^^^^^^
    * The famos BTTV driver has it implemented with a struct video_channel argument
    * and we follow it for compatibility reasons
    *
    * BTW: this is the only way the user can set the norm!
    */

   case VIDIOCSCHAN:
      {
         struct video_channel v;
         int input;
         int on, res;

         if (copy_from_user(&v, arg, sizeof(v))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCSCHAN: channel=%d, norm=%d\n",v.channel,v.norm);)
         switch (v.channel) {
         case 0:
            input = 3;
            break;
         case 1:
            input = 7;
            break;
         default:
            printk(MV300_ERR ": VIDIOCSCHAN on not existing channel %d\n",v.channel);
            return -EINVAL;
         }

         if (v.norm != VIDEO_MODE_PAL
          && v.norm != VIDEO_MODE_NTSC) {
            printk(MV300_ERR ": VIDIOCSCHAN with not supported norm %d\n",v.norm);
            return -EOPNOTSUPP;
         }
         btv->params.norm  = v.norm;
         btv->params.input = v.channel;

         /* We switch overlay off and on since a change in the norm
            needs different VFE settings */

         on = btv->v4l_overlay_active && !btv->v4l_memgrab_active;
         if (on) zr36057_overlay(btv, 0);

         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &input);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_NORM, &btv->params.norm);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_NORM, &btv->params.norm);

         if (on) zr36057_overlay(btv, 1);

         /* Make sure the changes come into effect */
         res = wait_grab_pending(btv);
         if(res) return res;

         return 0;
      }

   case VIDIOCGTUNER:
      {
         printk(MV300_ERR ": ioctl VIDIOCGTUNER not supported\n");
         return -EINVAL;
      }

   case VIDIOCSTUNER:
      {
         printk(MV300_ERR ": ioctl VIDIOCSTUNER not supported\n");
         return -EINVAL;
      }

   case VIDIOCGPICT:
      {
         struct video_picture p = btv->picture;

         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGPICT\n");)
         p.depth = btv->buffer.depth;
         switch (btv->buffer.depth) {
         case 15:
            p.palette = VIDEO_PALETTE_RGB555;
            break;

         case 16:
            p.palette = VIDEO_PALETTE_RGB565;
            break;

         case 24:
            p.palette = VIDEO_PALETTE_RGB24;
            break;

         case 32:
            p.palette = VIDEO_PALETTE_RGB32;
            break;
         }

         if (copy_to_user(arg, &p, sizeof(p))) {
            return -EFAULT;
         }
         return 0;
      }

   case VIDIOCSPICT:
      {
         struct video_picture p;

         if (copy_from_user(&p, arg, sizeof(p))) {
            return -EFAULT;
         }
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_PICTURE, &p);
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCSPICT bri=%d hue=%d col=%d con=%d dep=%d pal=%d\n",
                            p.brightness,p.hue,p.colour,p.contrast,p.depth,p.palette);)
         /* The depth and palette values have no meaning to us,
            should we return  -EINVAL if they don't fit ? */
         btv->picture = p;
         return 0;
      }

   case VIDIOCCAPTURE:
      {
         int v, res;

         if (copy_from_user(&v, arg, sizeof(v))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCCAPTURE: %d\n",v);)

         /* If there is nothing to do, return immediatly */

         if ( (v && btv->v4l_overlay_active) || (!v && !btv->v4l_overlay_active) ) return 0;

         if (v==0)
         {
            btv->v4l_overlay_active = 0;
            if (!btv->v4l_memgrab_active) zr36057_overlay(btv, 0);
            /* When a grab is running, the video simply won't be switched on any more */
         }
         else
         {
            if (!btv->buffer_set || !btv->window_set) {
               printk(MV300_ERR ": VIDIOCCAPTURE: buffer or window not set\n");
               return -EINVAL;
            }
            btv->v4l_overlay_active = 1;
            if (!btv->v4l_memgrab_active) zr36057_overlay(btv, 1);
            /* When a grab is running, the video will be switched on when grab is finished */
         }
         /* Make sure the changes come into effect */
         res = wait_grab_pending(btv);
         if(res) return res;
         return 0;
      }

   case VIDIOCGWIN:
      {
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGWIN\n");)
         if (copy_to_user(arg, &btv->window, sizeof(btv->window))) {
            return -EFAULT;
         }
         return 0;
      }

   case VIDIOCSWIN:
      {
         struct video_clip * vcp;
         struct video_window vw;
         int on, end, res;

         if (copy_from_user(&vw, arg, sizeof(vw))) {
            return -EFAULT;
         }

         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCSWIN: x=%d y=%d w=%d h=%d clipcount=%d\n",
                            vw.x,vw.y,vw.width,vw.height,vw.clipcount);)

         if (!btv->buffer_set) {
            printk(MV300_ERR ": VIDIOCSWIN: frame buffer has to be set first\n");
            return -EINVAL;
         }

         /*
          * The video front end needs 4-byte alinged line sizes, we correct that
          * silently here if necessary
          */

         if(btv->buffer.depth==15 || btv->buffer.depth==16)
         {
            end  = (vw.x+vw.width) & ~1; /* round down */
            vw.x = (vw.x+1) & ~1;        /* round up */
            vw.width = end-vw.x;
         }

         if(btv->buffer.depth==24)
         {
            end  = (vw.x+vw.width) & ~3; /* round down */
            vw.x = (vw.x+3) & ~3;        /* round up */
            vw.width = end-vw.x;
         }

#if 0
         // At least xawtv seems to care about the following - just leave it away
         /*
          * Also corrected silently (as long as window fits at all):
          * video not fitting the screen
          */
#if 0
         if (vw.x<0 || vw.y<0 || vw.x+vw.width>btv->buffer.width ||
                                 vw.y+vw.height>btv->buffer.height) {
            printk(MV300_ERR ": VIDIOCSWIN: window does not fit frame buffer: %dx%d+%d*%d\n",
                              vw.width,vw.height,vw.x,vw.y);
            return -EINVAL;
         }
#else
         if(vw.x           < 0                 ) vw.x      = 0;
         if(vw.y           < 0                 ) vw.y      = 0;
         if(vw.x+vw.width  > btv->buffer.width ) vw.width  = btv->buffer.width  - vw.x;
         if(vw.y+vw.height > btv->buffer.height) vw.height = btv->buffer.height - vw.y;
#endif
#endif

         /* Check for vaild parameters */
         if (vw.width<BUZ_MIN_WIDTH || vw.height<BUZ_MIN_HEIGHT ||
             vw.width>BUZ_MAX_WIDTH || vw.height>BUZ_MAX_HEIGHT ) {
            printk(MV300_ERR ": VIDIOCSWIN: width = %d or height = %d invalid\n",vw.width,vw.height);
            return -EINVAL;
         }

#ifdef XAWTV_HACK
         if(vw.width>720) vw.width=720;
#endif
  
         btv->window.x         = vw.x;
         btv->window.y         = vw.y;
         btv->window.width     = vw.width;
         btv->window.height    = vw.height;
         btv->window.chromakey = 0;
         btv->window.flags     = 0; // RJ: Is this intended for interlace on/off ?
         btv->window.clips     = NULL;
         btv->window.clipcount = vw.clipcount;

         /*
          * If an overlay is running, we have to switch it off
          * and switch it on again in order to get the new settings in effect.
          *
          * We also want to avoid that the overlay mask is written
          * when an overlay is running.
          */

         on = btv->v4l_overlay_active && !btv->v4l_memgrab_active;
         if (on) zr36057_overlay(btv, 0);

         /*
          *   Write the overlay mask if clips are wanted.
          */
         if (vw.clipcount) {
            vcp = vmalloc(sizeof(struct video_clip) * (vw.clipcount + 4));
            if (vcp == NULL) {
               printk(MV300_ERR ": mv300_ioctl: Alloc of clip mask failed\n");
               return -ENOMEM;
            }
            if (copy_from_user(vcp, vw.clips, sizeof(struct video_clip) * vw.clipcount)) {
               vfree(vcp);
               return -EFAULT;
            }
            write_overlay_mask(btv, vcp, vw.clipcount);
            vfree(vcp);
         }

         if (on) zr36057_overlay(btv, 1);
         btv->window_set = 1;

         /* Make sure the changes come into effect */
         res = wait_grab_pending(btv);
         if(res) return res;

         return 0;
      }

   case VIDIOCGFBUF:
      {
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGFBUF\n");)
         if (copy_to_user(arg, &btv->buffer, sizeof(btv->buffer))) {
            return -EFAULT;
         }
         return 0;
      }

   case VIDIOCSFBUF:
      {
         struct video_buffer v;

         /* RJ: Isn't this too restrictive? As long as the user doesn't set
                the base address it shouldn't be too dangerous */

         if (!capable(CAP_SYS_ADMIN)) {
            printk(MV300_ERR ": Only the superuser may issue VIDIOCSFBUF ioctl\n");
            return -EPERM;
         }
         if (copy_from_user(&v, arg, sizeof(v))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCSFBUF: base=0x%x w=%d h=%d depth=%d bpl=%d\n",
                            (u32)v.base,v.width,v.height,v.depth,v.bytesperline);)
         if ( btv->v4l_overlay_active ) {
            /* Has the user gotten crazy ... ? */
            printk(MV300_ERR ": VIDIOCSFBUF not allowed when overlay active\n");
            return -EINVAL;
         }
         if (  v.depth != 15
            && v.depth != 16
            && v.depth != 24
            && v.depth != 32) {
            printk(MV300_ERR ": VIDIOCSFBUF: depth=%d not supported\n",v.depth);
            return -EINVAL;
         }
         if (v.height<=0 || v.width <=0 || v.bytesperline <= 0) {
            printk(MV300_ERR ": VIDIOCSFBUF: invalid height/width/bpl value\n");
            return -EINVAL;
         }
         if (v.bytesperline&3) {
            printk(MV300_ERR ": VIDIOCSFBUF: bytesperline must be 4-byte aligned\n");
            return -EINVAL;
         }
         if (v.base) {
            btv->buffer.base = (void*)((unsigned long)v.base & ~3);
         }
         btv->buffer.height       = v.height;
         btv->buffer.width        = v.width;
         btv->buffer.depth        = v.depth;
         btv->buffer.bytesperline = v.bytesperline;

         if(btv->buffer.base) btv->buffer_set = 1;
         btv->window_set = 0; /* The user should set new window parameters */
         return 0;
      }

   /* RJ: what is VIDIOCKEY intended to do ??? */

   case VIDIOCKEY:
      {
         /* Will be handled higher up .. */
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCKEY\n");)
         return 0;
      }

   case VIDIOCGFREQ:
      {
         printk(MV300_ERR ": ioctl VIDIOCGFREQ not supported\n");
         return -EINVAL;
      }

   case VIDIOCSFREQ:
      {
         printk(MV300_ERR ": ioctl VIDIOCSFREQ not supported\n");
         return -EINVAL;
      }

   case VIDIOCGAUDIO:
      {
         printk(MV300_ERR ": ioctl VIDIOCGAUDIO not supported\n");
         return -EINVAL;
      }

   case VIDIOCSAUDIO:
      {
         printk(MV300_ERR ": ioctl VIDIOCSAUDIO not supported\n");
         return -EINVAL;
      }

   case VIDIOCSYNC:
      {
         int v;

         if (copy_from_user(&v, arg, sizeof(v))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCSYNC %d\n",v);)
         return v4l_sync(btv, v);
      }

   case VIDIOCMCAPTURE:
      {
         struct video_mmap vm;

         if (copy_from_user((void *) &vm, (void *)arg, sizeof(vm))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCMCAPTURE frame=%d geom=%dx%d fmt=%d\n",
                            vm.frame,vm.height,vm.width,vm.format);)
         return v4l_grab(btv, &vm);
      }

   case VIDIOCGMBUF:
      {
         struct video_mbuf vm;
         int i;

         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGMBUF\n");)
         vm.size   = v4l_nbufs * v4l_bufsize;
         vm.frames = v4l_nbufs;
         for (i = 0; i < v4l_nbufs; i++) {
            vm.offsets[i] = i * v4l_bufsize;
         }

         /* The next mmap will map the V4L buffers */
         btv->map_mjpeg_buffers = 0;

         if (copy_to_user(arg, &vm, sizeof(vm))) {
            return -EFAULT;
         }
         return 0;
      }

   case VIDIOCGUNIT:
      {
         struct video_unit vu;

         IOCTL_DEBUG(printk("mv300 ioctl VIDIOCGUNIT\n");)
         vu.video    = btv->video_dev.minor;
         vu.vbi      = VIDEO_NO_UNIT;
         vu.radio    = VIDEO_NO_UNIT;
         vu.audio    = VIDEO_NO_UNIT;
         vu.teletext = VIDEO_NO_UNIT;
         if (copy_to_user(arg, &vu, sizeof(vu))) {
            return -EFAULT;
         }
         return 0;
      }

   /*
    * RJ: In principal we could support subcaptures for V4L grabbing.
    *     Not even the famous BTTV driver has them, however.
    *     If there should be a strong demand, one could consider
    *     to implement them.
    */
   case VIDIOCGCAPTURE:
      {
         printk(MV300_ERR ": ioctl VIDIOCGCAPTURE not supported\n");
         return -EINVAL;
      }

   case VIDIOCSCAPTURE:
      {
         printk(MV300_ERR ": ioctl VIDIOCSCAPTURE not supported\n");
         return -EINVAL;
      }

   case BUZIOC_G_PARAMS:
      {
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_G_PARAMS\n");)

         if (copy_to_user(arg, &(btv->params), sizeof(btv->params))) {
            return -EFAULT;
         }
         return 0;
      }

   case BUZIOC_S_PARAMS:
      {
         struct buz_params bp;
         int input, on;

         if( btv->codec_mode != BUZ_MODE_IDLE) {
            printk(MV300_ERR ": BUZIOC_S_PARAMS called but mv300 in capture/playback mode\n");
            return -EINVAL;
         }

         if (copy_from_user(&bp, arg, sizeof(bp))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_S_PARAMS\n");)

         /* Check the params first before overwriting our internal values */

         if(mv300_check_params(&bp)) return -EINVAL;

         btv->params = bp;

         /* Make changes of input and norm go into effect immediatly */

         /* We switch overlay off and on since a change in the norm
            needs different VFE settings */

         on = btv->v4l_overlay_active && !btv->v4l_memgrab_active;
         if (on) zr36057_overlay(btv, 0);

         input = btv->params.input==0 ? 0 : 1;
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &input);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_NORM, &btv->params.norm);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_NORM, &btv->params.norm);

         if (on) zr36057_overlay(btv, 1);

         if (copy_to_user(arg, &bp, sizeof(bp))) {
            return -EFAULT;
         }
         return 0;
      }

   case BUZIOC_REQBUFS:
      {
         struct buz_requestbuffers br;

         if(btv->jpg_buffers_allocated) {
            printk(MV300_ERR ": BUZIOC_REQBUFS: buffers allready allocated\n");
            return -EINVAL;
         }
         if (copy_from_user(&br, arg, sizeof(br))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_REQBUFS count = %lu size=%lu\n",
                            br.count,br.size);)

         /* Enforce reasonable lower and upper limits */
         if(br.count<4)             br.count = 4; /* Could be choosen smaller */
         if(br.count>BUZ_MAX_FRAME) br.count = BUZ_MAX_FRAME;
         br.size = PAGE_ALIGN(br.size);
         if(br.size < 8192)       br.size = 8192; /* Arbitrary */
         /* br.size is limited by 1 page for the stat_com tables to a Maximum of 2 MB */
         if(br.size > (600*1024)) br.size = (600*1024); /* 512 K should be enough */
         if(btv->need_contiguous && br.size>MAX_KMALLOC_MEM) br.size = MAX_KMALLOC_MEM;

         btv->jpg_nbufs   = br.count;
         btv->jpg_bufsize = br.size;

         if(jpg_fbuffer_alloc(btv)) return -ENOMEM;

         /* The next mmap will map the MJPEG buffers */
         btv->map_mjpeg_buffers = 1;

         if (copy_to_user(arg, &br, sizeof(br))) {
            return -EFAULT;
         }
         return 0;
      }

   case BUZIOC_QBUF_CAPT:
      {
         int nb;

         if (copy_from_user((void *)&nb, (void *)arg, sizeof(int))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_QBUF_CAPT %d\n",nb);)
         return jpg_qbuf(btv, nb, BUZ_MODE_MOTION_COMPRESS);
      }

   case BUZIOC_QBUF_PLAY:
      {
         int nb;

         if (copy_from_user((void *)&nb, (void *)arg, sizeof(int))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_QBUF_PLAY %d\n",nb);)
         return jpg_qbuf(btv, nb, BUZ_MODE_MOTION_DECOMPRESS);
      }

   case BUZIOC_SYNC:
      {
         struct buz_sync bs;
         int res;

         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_SYNC\n");)
         res = jpg_sync(btv, &bs);
         if (copy_to_user(arg, &bs, sizeof(bs))) {
            return -EFAULT;
         }
         return res;
      }

   case BUZIOC_G_STATUS:
      {
         struct buz_status bs;
         int norm, input, status;
         unsigned long timeout;

         if( btv->codec_mode != BUZ_MODE_IDLE) {
            printk(MV300_ERR ": BUZIOC_G_STATUS called but mv300 in capture/playback mode\n");
            return -EINVAL;
         }

         if (copy_from_user(&bs, arg, sizeof(bs))) {
            return -EFAULT;
         }
         IOCTL_DEBUG(printk("mv300 ioctl BUZIOC_G_STATUS\n");)

         switch (bs.input) {
         case 0:
            input = 0;
            break;
         case 1:
            input = 1;
            break;
         default:
            printk(MV300_ERR ": BUZIOC_G_STATUS on not existing input %d\n",bs.input);
            return -EINVAL;
         }

         /* Set video norm to VIDEO_MODE_AUTO */

         norm = VIDEO_MODE_AUTO;
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT,  &input);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_NORM,   &norm   );

         /* sleep 1 second */

         timeout = jiffies + 1*HZ;
         while(jiffies<timeout) schedule();

         /* Get status of video decoder */

         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_GET_STATUS, &status );
         bs.signal = (status&DECODER_STATUS_GOOD)  ? 1 : 0;
         bs.norm   = (status&DECODER_STATUS_NTSC)  ? VIDEO_MODE_NTSC : VIDEO_MODE_PAL;
         bs.color  = (status&DECODER_STATUS_COLOR) ? 1 : 0;

         /* restore previous input and norm */
         input = btv->params.input==0 ? 0 : 1;
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &input);
         i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_NORM, &btv->params.norm);

         if (copy_to_user(arg, &bs, sizeof(bs))) {
            return -EFAULT;
         }
         return 0;
      }

   default:
      IOCTL_DEBUG(printk("mv300 ioctl UNKNOWN ioctl cmd: 0x%x\n",cmd);)
      return -ENOIOCTLCMD;

   }
   return 0;
}


/*
 *   This maps the buffers to user space.
 *
 *   Depending on the state of btv->map_mjpeg_buffers
 *   the V4L or the MJPEG buffers are mapped
 *
 */

static int mv300_mmap(struct video_device * dev, const char * adr, unsigned long size)
{
   struct buz * btv = (struct buz *)dev;
   unsigned long start = (unsigned long) adr;
   unsigned long page, pos, todo, fraglen;
   int i, j;

   if ( btv->map_mjpeg_buffers)
   {
      /* Map the MJPEG buffers */

      if(!btv->jpg_buffers_allocated)
      {
         printk(MV300_ERR ": mv300_mmap(MJPEG): buffers not yet allocated\n");
         return -ENOMEM;
      }

      if(size > btv->jpg_nbufs*btv->jpg_bufsize)
      {
         printk(MV300_ERR ": mv300_mmap(MJPEG): Max size is %lu - you wanted %lu\n",
                btv->jpg_nbufs*btv->jpg_bufsize, size);
         return -EINVAL;
      }

      if(size != btv->jpg_nbufs*btv->jpg_bufsize)
         printk(MV300_WARNING ": mv300_mmap(MJPEG): Expected %lu - you wanted %lu\n",
                btv->jpg_nbufs*btv->jpg_bufsize, size);

      for(i=0;i<btv->jpg_nbufs;i++)
      {
         for(j=0;j<btv->jpg_bufsize/PAGE_SIZE;j++)
         {
            fraglen = (btv->jpg_gbuf[i].frag_tab[2*j+1] & ~1) << 1;
            todo = size;
            if(todo>fraglen) todo = fraglen;
            pos = (unsigned long)btv->jpg_gbuf[i].frag_tab[2*j];
            page = virt_to_phys(bus_to_virt(pos)); /* should just be pos on i386 */
            if (remap_page_range(start, page, todo, PAGE_SHARED)) {
               printk(MV300_ERR ": mv300_mmap(V4L): remap_page_range failed\n");
               return -EAGAIN;
            }
            size  -= todo;
            start += todo;
            if(size==0) break;
            if(btv->jpg_gbuf[i].frag_tab[2*j+1] & 1) break; /* was last fragment */
         }
         if(size==0) break;
      }
   }
   else
   {
      /* Map the V4L buffers */

      if(size > v4l_nbufs*v4l_bufsize)
      {
         printk(MV300_ERR ": mv300_mmap(V4L): Max size is %d - you wanted %ld\n",
                v4l_nbufs*v4l_bufsize, size);
         return -EINVAL;
      }

      if(size != v4l_nbufs*v4l_bufsize)
         printk(MV300_WARNING ": mv300_mmap(V4L): Expected %d - you wanted %ld\n",
                v4l_nbufs*v4l_bufsize, size);

      for(i=0;i<v4l_nbufs;i++)
      {
         todo = size;
         if(todo>v4l_bufsize) todo = v4l_bufsize;
         page = btv->v4l_gbuf[i].fbuffer_phys;
         DEBUG(printk("V4L remap page range %d 0x%x %d to 0x%x\n",i,page,todo,start);)
         if (remap_page_range(start, page, todo, PAGE_SHARED)) {
            printk(MV300_ERR ": mv300_mmap(V4L): remap_page_range failed\n");
            return -EAGAIN;
         }
         size  -= todo;
         start += todo;
         if(size==0) break;
      }
   }
   return 0;
}

static int mv300_init_done(struct video_device * dev)
{
   return 0;
}

static struct video_device mv300_template = {
   MV300_NAME,
   VID_TYPE_CAPTURE | VID_TYPE_OVERLAY | VID_TYPE_CLIPPING | VID_TYPE_FRAMERAM |
   VID_TYPE_SCALES | VID_TYPE_SUBCAPTURE,
   VID_HARDWARE_BT848, /* Not true, but the mv300 is not yet in the list */
   mv300_open,
   mv300_close,
   mv300_read,
   mv300_write,
   NULL,
   mv300_ioctl,
   mv300_mmap,
   mv300_init_done,
   NULL,
   0,
   0
};

static int zr36057_init(int i)
{
   struct buz * btv = &mv300s[i];
   unsigned long mem;
   unsigned mem_needed;
   int j;

   /* reset zr36057 */
   btwrite(0, ZR36057_SPGPPCR);
   mdelay(10);

   /* default setup of all parameters which will persist beetween opens */

   btv->user = 0;

   btv->map_mjpeg_buffers = 0; /* Map V4L buffers by default */

   btv->jpg_nbufs   = 0;
   btv->jpg_bufsize = 0;
   btv->jpg_buffers_allocated = 0;

   btv->buffer_set    = 0;     /* Flag if frame buffer has been set */
   btv->buffer.base   = (void *)vidmem;
   btv->buffer.width  = 0;
   btv->buffer.height = 0;
   btv->buffer.depth  = 0;
   btv->buffer.bytesperline = 0;

   btv->params.norm     = default_norm  ? 1 : 0; /* Avoid nonsense settings from user */
   btv->params.input    = default_input ? 1 : 0; /* Avoid nonsense settings from user */
   btv->video_interlace = 0;

   /* Should the following be reset at every open ? */

   btv->picture.colour     = 32768;
   btv->picture.brightness = 32768;
   btv->picture.hue        = 32768;
   btv->picture.contrast   = 32768;
   btv->picture.whiteness  = 0;
   btv->picture.depth      = 0;
   btv->picture.palette    = 0;

   for(j=0;j<VIDEO_MAX_FRAME;j++)
   {
      btv->v4l_gbuf[i].fbuffer      = 0;
      btv->v4l_gbuf[i].fbuffer_phys = 0;
      btv->v4l_gbuf[i].fbuffer_bus  = 0;
   }

   btv->stat_com = 0;

   /* default setup (will be repeated at every open) */

   mv300_open_init_params(btv);

   /* allocate memory *before* doing anything to the hardware in case allocation fails */

   /* STAT_COM table and overlay mask */

   mem_needed = (BUZ_NUM_STAT_COM + ((BUZ_MAX_WIDTH + 31) / 32) * BUZ_MAX_HEIGHT)*4;
   mem = (unsigned long) kmalloc(mem_needed, GFP_KERNEL);
   if (!mem) {
      printk(MV300_ERR "-%u zr36057_init: kmalloc (STAT_COM + ovl.mask) failed\n",btv->id);
      return -ENOMEM;
   }
   memset((void*)mem, 0, mem_needed);

   btv->stat_com = (u32*)mem;
   for (j = 0; j < BUZ_NUM_STAT_COM; j++) {
      btv->stat_com[j] = 1;      /* mark as unavailable to zr36057 */
   }
   btv->overlay_mask = (u32 *)(mem + BUZ_NUM_STAT_COM*4);

   /* Initialize btv->jpg_gbuf */

   for (j = 0; j < BUZ_MAX_FRAME; j++) {
      btv->jpg_gbuf[j].frag_tab       = 0;
      btv->jpg_gbuf[j].frag_tab_bus   = 0;
      btv->jpg_gbuf[j].state          = BUZ_STATE_USER;
      btv->jpg_gbuf[j].bs.frame       = j;
   }

   /* take zr36057 out of reset now */
   btwrite(ZR36057_SPGPPCR_SoftReset, ZR36057_SPGPPCR);
   mdelay(10);

   /* stop all DMA processes */
   btwrite(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
   btand(~ZR36057_VDCR_VidEn, ZR36057_VDCR);
   /* assert P_Reset */
   btwrite(0, ZR36057_JPC);

   /* set up GPIO direction */
   btwrite(ZR36057_SPGPPCR_SoftReset | 0x00, ZR36057_SPGPPCR);

   /* Set up guest bus timing - Guests 0..3 Tdur=12, Trec=3 */
   btwrite((GPIO_MASK << 24) | 0x8888, ZR36057_GPPGCR1);
   mdelay(10);

   /* reset video decoder, encoder and codec */
   GPIO(btv, 0, 1);
   mdelay(10);
   GPIO(btv, 0, 0);
   mdelay(10);
   GPIO(btv, 0, 1);
   mdelay(10);

   /* wakeup video encoder and set slave */
   GPIO(btv, 4, 0);
   mdelay(10);
   GPIO(btv, 2, 1);
   mdelay(10);

   /* Set ZR36060 FRAME pin Low */
   GPIO(btv, 3, 1);
   mdelay(10);

   /* Set All Rest GPIO Pin High */
   GPIO(btv, 5, 1);
   mdelay(10);
   GPIO(btv, 6, 1);
   mdelay(10);
   GPIO(btv, 7, 1);
   mdelay(10);

   /* reset JPEG codec */
   zr36060_sleep(btv, 0);
   mdelay(10);
   zr36060_reset(btv);
   mdelay(10);

   /* display codec revision */
   if (zr36060_read_8(btv, 0x022) == 0x33) {
      printk(MV300_INFO "-%u: Zoran ZR36060 (rev %d)\n",
             btv->id, zr36060_read_8(btv, 0x023));
   } else {
      printk(MV300_ERR "-%u: Zoran ZR36060 not found\n", btv->id);
      kfree((void *)btv->stat_com);
      return -1;
   }

   /* i2c */
   memcpy(&btv->i2c, &mv300_i2c_bus_template, sizeof(struct i2c_bus));
   sprintf(btv->i2c.name, MV300_NAME "-%u", btv->id);
   btv->i2c.data = btv;
   if (i2c_register_bus(&btv->i2c) < 0) {
      kfree((void *)btv->stat_com);
      return -1;
   }

   /*
    *   Now add the template and register the device unit.
    */
   memcpy(&btv->video_dev, &mv300_template, sizeof(mv300_template));
   sprintf(btv->video_dev.name, MV300_NAME "-%u", btv->id);
   if (video_register_device(&btv->video_dev, VFL_TYPE_GRABBER) < 0) {
      i2c_unregister_bus(&btv->i2c);
      kfree((void *)btv->stat_com);
      return -1;
   }

   /* toggle JPEG codec sleep to sync PLL */
   zr36060_sleep(btv, 1);
   mdelay(10);
   zr36060_sleep(btv, 0);
   mdelay(10);

   /* Enable bus-mastering */
   pci_set_master(btv->pci_dev);

   j = btv->params.input==0 ? 3 : 7;
   i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &j);
   i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_NORM, &btv->params.norm);
   i2c_control_device(&btv->i2c, I2C_DRIVERID_ENCODER, ENCODER_SET_NORM, &btv->params.norm);

   /* set individual interrupt enables (without GIRQ0)
      but don't global enable until mv300_open() */

   btwrite(IRQ_MASK & ~ZR36057_ISR_GIRQ0, ZR36057_ICR);

   btv->initialized = 1;
   return 0;
}



static void release_mv300(void)
{
   u8 command;
   int i;
   struct buz * btv;

   for (i = 0; i < mv300_num; i++) {
      btv = &mv300s[i];

      if (!btv->initialized) continue;

      /* unregister i2c_bus */
      i2c_unregister_bus((&btv->i2c));

      /* disable PCI bus-mastering */
      pci_read_config_byte(btv->pci_dev, PCI_COMMAND, &command);
      command &= ~PCI_COMMAND_MASTER;
      pci_write_config_byte(btv->pci_dev, PCI_COMMAND, command);

      /* put chip into reset */
      btwrite(0, ZR36057_SPGPPCR);

      free_irq(btv->pci_dev->irq, btv);

      /* unmap and free memory */

      kfree((void *)btv->stat_com);

      iounmap(btv->zr36057_mem);

      video_unregister_device(&btv->video_dev);
   }
}

/*
 *   Scan for a MV300 card (actually for the PCI contoler ZR36057/67),
 *   request the irq and map the io memory
 */

static int find_zr36057(void)
{
   unsigned char latency;
   struct buz * btv;
   struct pci_dev * dev = NULL;
   int result;

   mv300_num = 0;

   while ( mv300_num < MV300_MAX
      &&  (dev = pci_find_device(PCI_VENDOR_ID_ZORAN, PCI_DEVICE_ID_ZORAN_36057, dev)) != NULL) {
      btv = &mv300s[mv300_num];
      btv->pci_dev = dev;
      btv->zr36057_mem = NULL;
      btv->id = mv300_num;
      sprintf(btv->name, MV300_NAME "-%u", btv->id);

      spin_lock_init(&btv->lock);

      btv->zr36057_adr = btv->pci_dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
      pci_read_config_byte(btv->pci_dev, PCI_CLASS_REVISION, &btv->revision);
      if (btv->revision < 2) {
         printk(MV300_INFO "-%u: Zoran ZR36057 (rev %d) irq: %d, memory: 0x%08x.\n",
                btv->id, btv->revision, btv->pci_dev->irq, btv->zr36057_adr);
      } else {
         unsigned short ss_vendor_id, ss_id;

         pci_read_config_word(btv->pci_dev, PCI_SUBSYSTEM_VENDOR_ID, &ss_vendor_id);
         pci_read_config_word(btv->pci_dev, PCI_SUBSYSTEM_ID, &ss_id);
         printk(MV300_INFO "-%u: Zoran ZR36067 (rev %d) irq: %d, memory: 0x%08x\n",
                   btv->id, btv->revision, btv->pci_dev->irq, btv->zr36057_adr);
         printk(MV300_INFO "-%u: subsystem vendor=0x%04x id=0x%04x\n",
                btv->id, ss_vendor_id, ss_id);
      }

      btv->zr36057_mem = ioremap(btv->zr36057_adr, 0x1000);

      result = request_irq(btv->pci_dev->irq, mv300_irq,
                 SA_SHIRQ | SA_INTERRUPT, btv->name, (void *)btv);
      if (result < 0) {
         if (result == -EINVAL) {
            printk(MV300_ERR "-%u: Bad irq number or handler\n", btv->id);
         } else if (result == -EBUSY) {
            printk(MV300_ERR "-%u: IRQ %d busy, change your PnP config in BIOS\n",
                   btv->id, btv->pci_dev->irq);
         } else {
            printk(MV300_ERR "-%u: Can't assign irq, code %d\n", btv->id, result);
         }
         iounmap(btv->zr36057_mem);
         continue;
      }

      /* set PCI latency timer */
      pci_read_config_byte(btv->pci_dev, PCI_LATENCY_TIMER, &latency);
      if (latency != 48) {
         printk(MV300_INFO "-%u: Changing PCI latency from %d to 48.\n", btv->id, latency);
         latency = 48;
         pci_write_config_byte(btv->pci_dev, PCI_LATENCY_TIMER, latency);
      }

      mv300_num++;
   }
   if (mv300_num == 0) {
      printk(MV300_INFO ": no cards found.\n");
   }
   return mv300_num;
}

/* Scan for PCI display adapter
   if more than one card is present the last one is used for now */

#include "videocards.h"

static int find_vga(unsigned long * video_address)
{
   unsigned int class;
   unsigned short badr;
   int found = 0, i, tga_type;
   unsigned int vidadr = 0;
   struct pci_dev * dev;

   for (dev = pci_devices; dev != NULL; dev = dev->next) {
      class = dev->class >> 8;
      if ((class >> 8) == PCI_BASE_CLASS_DISPLAY ||
          /* Number 9 GXE64Pro needs this */
          class == PCI_CLASS_NOT_DEFINED_VGA) {
         badr = 0;
         printk(MV300_INFO ": Found PCI display adapter: ");
         for (i = 0; i < NR_CARDS; i++) {
            if (dev->vendor == vbs[i].vendor) {
               if (vbs[i].device
                   && vbs[i].device != dev->device) {
                  continue;
               }
               printk("%s.\n", vbs[i].name);
               badr = vbs[i].badr;
               break;
            }
         }
         if (!badr) {
            printk("Unknown video card.\n");
            continue;
         }
         pci_read_config_dword(dev, badr, &vidadr);
         if (vidadr & PCI_BASE_ADDRESS_SPACE_IO) {
            printk(MV300_ERR ": Memory seems to be I/O memory.\n");
            printk(MV300_ERR ": Check entry for your card type in videocards.h\n");
            continue;
         }
         vidadr &= PCI_BASE_ADDRESS_MEM_MASK;
         if (!vidadr) {
            printk(MV300_ERR ": Memory @ 0, must be something wrong!\n");
            continue;
         }

         if (dev->vendor == PCI_VENDOR_ID_DEC) {
            if (dev->device == PCI_DEVICE_ID_DEC_TGA) {
               tga_type = (readl((unsigned long)vidadr) >> 12) & 0x0f;
               if (tga_type != 0
                   && tga_type != 1
                   && tga_type != 3) {
                  printk(MV300_ERR ": TGA type (0x%x) unrecognized!\n", tga_type);
                  found--;
               }
               vidadr += dec_offsets[tga_type];
            }
         }

         DEBUG(printk(MV300_DEBUG ": memory @ 0x%08x, devfn: 0x%04x.\n",
                 vidadr, dev->devfn));
         found++;
      }
   }

   *video_address = vidadr;

   return found;
}

#include "chipsets.h"

static void handle_chipset(void)
{
   int index;
   struct pci_dev * dev = NULL;

   for (index = 0; index < sizeof(black) / sizeof(black[0]); index++) {
      if ((dev = pci_find_device(black[index].vendor, black[index].device, dev)) != NULL) {
         printk(MV300_INFO ": Host bridge: %s, ", black[index].name);
         switch (black[index].action) {

         case TRITON:
            printk("enabling Triton support.\n");
            triton = 1;
            break;

         case NATOMA:
            printk("enabling Natoma workaround.\n");
            natoma = 1;
            break;
         }
      }
   }
}

#ifdef MODULE
int init_module(void)
#else
int init_mv300_cards(struct video_init * unused)
#endif
{
   int i;


   printk(MV300_INFO ": ----------------------------- driver loaded\n");

   /* Look for MV-300 cards */

   if (find_zr36057() < 0) {
      return -EIO;
   }

   printk(MV300_INFO ": %d mv300 card(s) found\n",mv300_num);

   if(mv300_num == 0) return -ENXIO;

   /* check the parameters we have been given, adjust if necessary */

   if(v4l_nbufs < 0) v4l_nbufs = 0;
   if(v4l_nbufs > VIDEO_MAX_FRAME) v4l_nbufs = VIDEO_MAX_FRAME;
   /* The user specfies the in KB, we want them in byte (and page aligned) */
   v4l_bufsize = PAGE_ALIGN(v4l_bufsize*1024);
   if(v4l_bufsize < 32768) v4l_bufsize = 32768;
   /* 2 MB is arbitrary but sufficient for the maximum possible images */
   if(v4l_bufsize > 2048*1024) v4l_bufsize = 2048*1024;

   printk(MV300_INFO ": using %d V4L buffers of size %d KB\n",v4l_nbufs,v4l_bufsize>>10);

   /* Use parameter for vidmem or try to find a video card */

   if(vidmem)
   {
      printk(MV300_INFO ": Using supplied video memory base address @ 0x%lx\n",vidmem);
   }
   else
   {
      printk(MV300_INFO ": No video memory address supplied, trying to find a videocard\n");
      if(find_vga(&vidmem))
         printk(MV300_INFO ": Known video card found, video memory base @ 0x%lx\n",vidmem);
      else
         printk(MV300_INFO ": No kown video card found\n");
   }

   /* check if we have a Triton or Natome chipset */

   handle_chipset();

   /* take care of Natoma chipset and a revision 1 zr36057 */

   for (i = 0; i < mv300_num; i++) 
   {
      if ( natoma && mv300s[i].revision <= 1)
      {
         mv300s[i].need_contiguous = 1;
         printk(MV300_INFO "-%u: ZR36057/Natome bug, max. buffer size is 128K\n",mv300s[i].id);
      }
      else
      {
         mv300s[i].need_contiguous = 0;
      }
   }

   /* initialize the MV-300s */

   /* We have to know which ones must be released if an error occurs */ 
   for (i = 0; i < mv300_num; i++) mv300s[i].initialized = 0;

   for (i = 0; i < mv300_num; i++) {
      if (zr36057_init(i) < 0) {
         release_mv300();
         return -EIO;
      }
   }

   return 0;
}



#ifdef MODULE

void cleanup_module(void)
{
   release_mv300();
}

#endif
